/**
 * @file TvTileGfx.h - provides TV tile graphics access via Andre's PASM tile graphics driver
 *
 * Copyright (c) 2011, David Betz / Andre' LaMothe
 *
 * ----------------------------------------------------------------------------
 *                        TERMS OF USE: MIT License
 * ----------------------------------------------------------------------------
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 * ----------------------------------------------------------------------------
 */

#ifndef __TV_TILE_GFX_H__
#define __TV_TILE_GFX_H__

#include <stdint.h>

#define TV_TILE_GFX_ROW_CNT				24
#define TV_TILE_GFX_COL_CNT				32
#define TV_TILE_GFX_TILE_SIZE			(sizeof(uint16_t) * 8)
#define TV_TILE_GFX_C64_FONT_TILE_CNT	144
#define TV_TILE_GFX_C64_PALETTE_CNT		16

typedef struct {
	uint16_t *tile_map_base_ptr_parm;      // parm 0: pointer to tile map (32x24 array, row major, 2-bytes per entry)
	uint16_t *tile_bitmaps_base_ptr_parm;  // parm 1: pointer to the base of the bitmap data, each 8x8, that is, 16 longs each
	uint32_t *tile_palettes_base_ptr_parm; // parm 2: pointer to the base of the palettes array, each palette 1 long, or 4 bytes.
	uint32_t tile_map_control_parm;        // parm 3: pointer to the value that holds the overscan colors and number of "screens" or multiples of 16 that the tile map is
	uint32_t tile_status_bits_parm;		   // parm 4: pointer to status word - vsync, hsync, etc.
	uint32_t video_vcfg;				   // parm 5: vcfg pingroup and pinmask fields
	uint32_t video_dira;				   // parm 6: dira value to enable pins
	int cogid;							   // number of cog running the driver + 1 or zero if the driver is not loaded
} TvTileGfx_st;

/* c64 font */
extern uint16_t c64_8x8_font[];
extern uint32_t c64_palettes[];
extern uint32_t c64_control;

/* Start TV Tile Graphics driver - starts a cog */
int TvTileGfx_start(TvTileGfx_st *data, int basepin);

/* Stop keyboard driver - frees a cog */
void TvTileGfx_stop(TvTileGfx_st *data);

#endif
