ifndef SRCDIR
SRCDIR=.
endif

ifndef HDRDIR
HDRDIR=.
endif

ifndef BOARD
BOARD=c3
endif

ifeq "$(BOARD)" "c3"
LDSCRIPT=c3.ld
CFLAGS+=-DBOARD_C3
endif

ifeq "$(BOARD)" "c3ram"
LDSCRIPT=c3ram.ld
CFLAGS+=-DBOARD_C3
endif

ifeq "$(BOARD)" "sdram"
LDSCRIPT=sdram.ld
CFLAGS+=-DBOARD_SDRAM
endif

ZOGHDRDIR=$(ZOGPATH)/include

OBJDIR=obj
BINDIR=bin
LIBDIR=$(ZOGPATH)/lib

CC=$(ZPUPATH)/zpu-elf-gcc
OC=$(ZPUPATH)/zpu-elf-objcopy

ECHO=echo
MKDIR=mkdir

CFLAGS+=-g -Wall -Os -I$(ZOGHDRDIR) -I$(HDRDIR)
LDFLAGS=-phi -Wl,-T$(LIBDIR)/$(LDSCRIPT) -Wl,--relax -Wl,--gc-sections -L$(LIBDIR)
OCFLAGS=-O binary -j .text -j .data -j .ctors -j .dtors

OBJS=$(foreach SRC,$(SOURCES),$(addsuffix .o,$(addprefix $(OBJDIR)/,$(basename $(SRC)))))

##################
# DEFAULT TARGET #
##################

.PHONY:	all
all:	$(TARGET)

###########################
# DOWNLOAD AND RUN TARGET #
###########################

.PHONY:	run
run:	$(TARGET)
	@$(ECHO) Loading through port $(PORT) to board $(BOARD)
	@$(ZOGPATH)/bin/$(OS)/zogload -p $(PORT) -b $(BOARD) -wm $(BINDIR)/$(TARGET).zbn -r -t

#################
# CLEAN TARGETS #
#################

.PHONY:	clean
clean:
	@rm -rf $(OBJDIR)
	@rm -rf $(BINDIR)

###############
# MAIN TARGET #
###############

ifneq "$(wildcard $(ZOGPATH)/libsrc)" ""
BUILDLIB=libsrc
endif

.PHONY:	$(TARGET)
$(TARGET):	$(BUILDLIB) $(BINDIR) $(OBJDIR) $(BINDIR)/$(TARGET).zbn

##################
# LIBRARY TARGET #
##################

.PHONY:	libsrc
libsrc:
	@$(MAKE) -C $(ZOGPATH)/libsrc

#########
# RULES #
#########

$(BINDIR)/$(TARGET).zbn:	$(BINDIR)/$(TARGET).elf
	@$(OC) $(OCFLAGS) $< $@
	@$(ECHO) $@

$(BINDIR)/$(TARGET).elf:	$(LIBDIR)/libprop.a $(OBJS)
	@$(CC) $(CFLAGS) $(LDFLAGS) $(OBJS) -o $@ $(LIBRARIES) -lprop
	@$(ECHO) $@

$(OBJDIR)/%.o:	$(SRCDIR)/%.c
	@$(CC) $(CFLAGS) -c $< -o $@
	@$(ECHO) $@

$(OBJDIR) $(BINDIR):
	$(MKDIR) $@
