#ifndef __PROPELLER_H__
#define __PROPELLER_H__

#ifdef __cplusplus
extern "C" 
{
#endif

#include <stdint.h>

#ifndef TRUE
#define TRUE    1
#define FALSE   0
#endif

#define hub_memory_base         0x10000000
#define hub_memory_size         32768
#define rendezvous_area_size    128     // make this the size of a cache line for alignment
#define vm_mbox_size            (2 * sizeof(uint32_t))
#define vm_cache_size           8192

#define rendezvous_area_off     (hub_memory_size - rendezvous_area_size)
#define vm_cache_off            (rendezvous_area_off - vm_cache_size)
#define vm_mbox_off             rendezvous_area_off

#define vm_mbox                 ((uint32_t *)(hub_memory_base + vm_mbox_off))

#define HUB						__attribute__ ((section(".hub")))
#define PhysicalHubAddress(a)   ((uint32_t)(a) - hub_memory_base)

extern uint8_t scratch_buffer[];
extern volatile uint32_t CLKFREQ;
extern int errno;

/* COG register definitions thanks to "jazzed" */

#define COGregBase 0x18000000

/**
 * define cog registers for C.
 * since registers are 32 bits, we have to shift them
 */
enum COGregAddrs_en
{
    _PAR  = COGregBase + (0x1f0<<2),
    _CNT  = COGregBase + (0x1f1<<2),
    _INA  = COGregBase + (0x1f2<<2),
    _INB  = COGregBase + (0x1f3<<2),
    _OUTA = COGregBase + (0x1f4<<2),
    _OUTB = COGregBase + (0x1f5<<2),
    _DIRA = COGregBase + (0x1f6<<2),
    _DIRB = COGregBase + (0x1f7<<2),
    _CTRA = COGregBase + (0x1f8<<2),
    _CTRB = COGregBase + (0x1f9<<2),
    _FRQA = COGregBase + (0x1fa<<2),
    _FRQB = COGregBase + (0x1fb<<2),
    _PHSA = COGregBase + (0x1fc<<2),
    _PHSB = COGregBase + (0x1fd<<2),
    _VCFG = COGregBase + (0x1fe<<2),
    _VSCL = COGregBase + (0x1ff<<2)
};

#define PAR  (*(volatile uint32_t *)_PAR)
#define CNT  (*(volatile uint32_t *)_CNT)
#define INA  (*(volatile uint32_t *)_INA)
#define INB  (*(volatile uint32_t *)_INB)
#define OUTA (*(volatile uint32_t *)_OUTA)
#define OUTB (*(volatile uint32_t *)_OUTB)
#define DIRA (*(volatile uint32_t *)_DIRA)
#define DIRB (*(volatile uint32_t *)_DIRB)
#define FRQA (*(volatile uint32_t *)_FRQA)
#define FRQB (*(volatile uint32_t *)_FRQB)
#define PHSA (*(volatile uint32_t *)_PHSA)
#define PHSB (*(volatile uint32_t *)_PHSB)
#define VCFG (*(volatile uint32_t *)_VCFG)
#define VSCL (*(volatile uint32_t *)_VSCL)

/* ZPU's libgloss uses syscall IDs 1 to 18, we stay well away from that range */
#define SYS_cogid   8000
#define SYS_coginit 8001
#define SYS_cogstop 8002
#define SYS_mboxcmd 8003

extern int _syscall(int *foo, int ID, ...);

void wait(int per);

void longfill(uint32_t *dst, uint32_t fill, int count);
void longmove(uint32_t *dst, uint32_t *src, int count);
void longmove_swapwords(uint32_t *dst, uint32_t *src, int count);

int cogid(void);
int cognew(uint32_t *code, int size, uint32_t *par);
int coginit(int cog, uint32_t *code, int size, uint32_t *par);
void cogstop(int cog);

#define TERM_IN		0x01
#define TERM_OUT	0x02
#define TERM_IO		(TERM_IN | TERM_OUT)

int InitSerialTerm(int rxpin, int txpin, int mode, int baudrate, int flags);
int InitKeyboardTerm(int dpin, int cpin);
int InitTvTerm(int tvpin);
int InitTv2Term(int tvpin);
int InitFileIO(void);
int InitC3FileIO(int retries);

int inbyte(void);
void outbyte(int ch);

extern int (*term_inbyte_ptr)(void);
extern void (*term_outbyte_ptr)(int ch);
extern int (*file_read_ptr)(int fd, char *buf, int nbytes);
extern int (*file_write_ptr)(int fd, char *buf, int nbytes);

// include board-specific definitions
#include "board.h"

#ifdef __cplusplus
}
#endif

#endif
