' /////////////////////////////////////////////////////////////////////////////
{
 TvTileGfx.spin
 VERSION: 1.0
 AUTHOR: Andre' LaMothe
 HACKED FOR ZOG BY: David Betz

 LAST MODIFIED: 1.29.11

 COMMENTS: Modified from Chameleon to be used in "stand alone" mode for the C3.

 Specs:
 - 32x24 tile map, each entry 2 bytes, format: [palette index:tile index]
 - 8x8 bitmap tiles, 4 colors each, array of 8 WORDs, each WORD represents 8 pixels, 16 bytes each tile, 4 LONGs
 - Course scrolling, horizontal and vertical, playfield width 32, 64, 128, 256
 - Fine scrolling on vertical

 Features:
 - Multiple tile maps 256
 - Multiple palettes 256
 - Bitmaps 256
 - All assets base pointers passed to engine each frame, thus "on the fly" changes can be made
 - Supports page flipping, double buffering by design.
 - Engine passes back vertical region and hsyn/vsync status as well as current line
}
' /////////////////////////////////////////////////////////////////////////////
' ASSEMBLY LANGUAGE HEL_GFX_DRIVER
' Engine 3 Design Notes
' Status: In progress
'
' Specs:
' - 32x24 tile map, each entry 2 bytes, format: [palette index:tile index]
' - 8x8 bitmap tiles, 4 colors each, array of 8 WORDs, each WORD represents 8 pixels, 16 bytes each tile, 4 LONGs
' - Course scrolling, horizontal and vertical, playfield width 32, 64, 128, 256
'
' Features:
' - Multiple tile maps 256
' - Multiple palettes 256
' - Bitmaps 256
' - All assets base pointers passed to engine each frame, thus "on the fly" changes can be made
' - Supports page flipping, double buffering by design.
' - Engine passes back vertical region and hsyn/vsync status as well as current line
' /////////////////////////////////////////////////////////////////////////////


CON





  FNTSC         = 3_579_545      ' NTSC color clock frequency in HZ


  'LNTSC         = (220*16)       ' NTSC color cycles per line (220-227) * 16
  'SNTSC         = (44*16)        ' NTSC color cycles per sync (39-44) * 16

  LNTSC = (227*16 + 1*4)          ' NTSC color cycles per line (220-228) * 16 +- epsilon
  SNTSC = (51*16 - 0)             ' NTSC color cycles per sync (39-54) * 16 +- epsilon

  VNTSC         = (LNTSC-SNTSC)  ' NTSC color cycles per active video * 16
  PNTSC256      = (VNTSC/16)     ' NTSC color cycles per "compressed/expanded" on screen pixel
                                 ' allows us to put more or less pixels on the screen, but
                                 ' remember NTSC is 224 visible pixels -- period, so when we display more
                                 ' than 224 per line then we are getting chroma distortion on pixel boundaries
                                 ' a more recommended method for cleaner graphics is about 180-190 pixels horizontally
                                 ' this way you don't overdrive the chroma bandwidth which limits colors to 224
                                 ' color clocks per screen
                                 ' we are going to leave clocks per pixel the same, but CHANGE the clocks per frame to 1/2 the value
                                 ' it was for the 16 pixel tiles, this way the timing stays the same, but the waitvid will need
                                 ' 2 - 8 pixel sets 2x as fast



' Chameleon settings
  'VIDEO_PINMASK    = %0111_0000     ' vcfg S = pinmask  (pin15 ->0111_0000<-pin8), only want upper nibble lower 3-bits
  'VIDEO_PINGROUP   = 1              ' vcfg D = pingroup (Chameleon uses group 1, pins 8-15)
  VIDEO_SETUP      = %0_11_1_11_000 ' vcfg I = controls overall setting, we want baseband video on bottom nibble, 2-bit color, enable chroma on broadcast & baseband
  VIDEO_CNTR_SETUP = %00001_111     ' pll internal routed to Video, PHSx+=FRQx (mode 1) + pll(16x)
                                    ' needn't set D,S fields since they set pin A/B I/Os, but mode 1 is internal, thus irrelvant

{

 ' HYDRA settings
  'VIDEO_PINMASK    = %0000_0111     ' vcfg S = pinmask  (pin31 ->0000_0111<-pin24), only want lower 3-bits
  'VIDEO_PINGROUP   = 3              ' vcfg D = pingroup (Hydra uses group 3, pins 24-31)
  VIDEO_SETUP      = %0_10_1_01_000 ' vcfg I = controls overall setting, we want baseband video on bottom nibble, 2-bit color, enable chroma on broadcast & baseband
  VIDEO_CNTR_SETUP = %00001_111     ' pll internal routed to Video, PHSx+=FRQx (mode 1) + pll(16x)
                                    ' needn't set D,S fields since they set pin A/B I/Os, but mode 1 is internal, thus irrelvant

}


' format of tile_status_bits_parm, only the Vsync status bit is updated
'
' byte 3 (unused)|byte 2 (line)|   byte 1 (tile postion)    |                     byte 0 (sync and region)      |
'|x x x x x x x x| line 8-bits | row 4 bits | column 4-bits |x x x x | region 2-bits | hsync 1-bit | vsync 1-bit|
'   b31..b24         b23..b16      b15..b12     b11..b8                    b3..b2          b1            b0
' Region 0=Top Overscan, 1=Active Video, 2=Bottom Overscan, 3=Vsync
' NOTE: In this version of the tile engine only VSYNC and REGION are valid

  ' tile engine status bits/masks
  TSB_VSYNC                     = %0000_00_01    ' Vsync
  TSB_HSYNC                     = %0000_00_10    ' Hsync

  ' regions indicate which parts of the vertical scan the raster is in, useful to time all rendering during the total blanking period which
  ' is any region other than "01"
  TSB_REGION_MASK               = %0000_11_00    ' region status bits bitmask

  TSB_REGION_TOP_OVERSCAN       = %0000_00_00    ' top overscan region (00)
  TSB_REGION_ACTIVE_VIDEO       = %0000_01_00    ' active video region (01)
  TSB_REGION_BOTT_OVERSCAN      = %0000_10_00    ' bottom overscan region (10)
  TSB_REGION_VSYNC              = %0000_11_00    ' vsync region (replicated in vsync bit)


  ' register indexes
  CLKFREQ_REG = 0                ' register address of global clock frequency

  ' debuging stuff
  DEBUG_LED_PORT_MASK = $00000001 ' debug LED is on I/O P0



PUB start(params) | cog, cogon

cogon := (cog := cognew(@HEL_GFX_Driver_Entry, params)) > 0

DAT
                org $000  ' set the code emission for COG add $000

' /////////////////////////////////////////////////////////////////////////////
' Entry point and Parameter Interface
' /////////////////////////////////////////////////////////////////////////////
{

there are 4 parameters in this engine, listed in order below
tile engine data structure pointers (can be changed in real-time by app!)
input parms

long tile_map_base_ptr_parm      ' parm 0: pointer to tile map (32x24 array, row major, 2-bytes per entry)
long tile_bitmaps_base_ptr_parm  ' parm 1: pointer to the base of the bitmap data, each 8x8, that is, 16 bytes each
long tile_palettes_base_ptr_parm ' parm 2: pointer to the base of the palettes array, each palette 1 long, or 4 bytes.
long tile_map_control_parm       ' parm 3: pointer to the value that holds the overscan colors and number of "screens" or multiples of 16 that the tile map is
                                 ' format $bc_tc_hv_ww, tc = top overscan color, bc= bottom overscan color, h=horizontal scroll (0..7), v=vertical scroll (0..7),
                                 ' ww = width, set for width 16 tiles, 0 = 32 tiles, 1 = 64 tiles, 2 = 128 tiles, 3 = 256 tiles, etc.
                                 ' eg. w=0 would be 16 wide (standard), 1 would be 32 tiles, 2 would be 64 tiles, etc.
                                 ' this allows multiscreen width playfields and thus large horizontal/vertical scrolling games
                                 ' note that the final size is always a power of 2 times 16

real-time engine status variables, these are updated in real time by the
tile engine itself, so they can be monitored outside in SPIN/ASM by game
output parms

long tile_status_bits_parm      ' parm 4: pointer to status word - vsync, hsync, etc.

pin configuration

long video_vcfg                 ' parm 5: vcfg pingroup and pinmask fields
long video_dira                 ' parm 6: dira value to enable pins

 format of tile_status_bits_parm, only the Vsync status bit is updated

 byte 3 (unused)|byte 2 (line)|   byte 1 (tile postion)    |                     byte 0 (sync and region)      |
|x x x x x x x x| line 8-bits | row 4 bits | column 4-bits |x x x x | region 2-bits | hsync 1-bit | vsync 1-bit|
   b31..b24         b23..b16      b15..b12     b11..b8                    b3..b2          b1            b0
 Region 0=Top Overscan, 1=Active Video, 2=Bottom Overscan, 3=Vsync
 NOTE: In this version of the tile engine only VSYNC and REGION are valid

format of a tile map:

DAT

              ' you place all your 32x24 tile maps here, you can have as many as you like, in real-time simply re-point the
              ' tile_map_base_ptr_parm to any time map and within 1 frame the tile map will update

              ' 32x24 WORDS each, (0..768 WORDs,1536 bytes per tile map) 2-BYTE tiles (msb)[palette_index | tile_index](lsb)
              ' 32x24 tile map, each tile is 2 bytes, there are a total of 256 tiles possible, and thus 256 palettes

              ' column     0      1      2      3      4      5      6      7      8      9     10     11     12     13     14     15      16    17      18    19     20     21     22     23     24     25     26     27     28     29     20     31
tile_map0     word      $02_01,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_01 ' row 0
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 1
              word      $00_00,$00_01,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 2
              word      $00_00,$00_00,$00_01,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 3
              word      $00_00,$00_00,$00_00,$00_00,$00_01,$00_00,$00_00,$00_00,$00_00,$00_01,$00_00,$00_00,$00_01,$00_00,$00_00,$00_01,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 4
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 5
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 6
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 7
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_01,$00_00,$00_00,$00_00,$00_00,$00_00,$00_01,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 8
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 9
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_01,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 10
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_01,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 11
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 12
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 13
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 14
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 15
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 16
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 17
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 18
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 19
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 20
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 21
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 22
              word      $00_01,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_01 ' row 23

Example: upper left corner tile @ (0, 0) value is $02_01, this means use palette 2 and tile bitmap 1.
Note: 1. you can have as many tile maps as you wish there is no limit, they can represent rooms, etc.
      2. if you use "mappy" tool then you must make some changes tothe mapwin.ini file, they are:

      maptype="LW1H1A1-0" <--this results in 1 byte per entry, and the format is a binary single array in the format
      importskip=0        <---use dumb import, don't modify anything, just bring the tiles into the tool from your bitmap

      width, height,
      index 0, index 1, ..... index 31,
      index 32, index 1, ..... index 63,
      .
      .
      index 736, index 177, ..... index 191

      as one single byte stream of course.

DAT

              ' tile bitmap memory, each tile 8x8 pixels, or 1 WORD by 8,
              ' 16-bytes each, also, note that they are mirrored right to left
              ' since the VSU streams from low to high bits, so your art must
              ' be reflected, we could remedy this in the engine, but for fun
              ' I leave it as a challenge in the art, since many engines have
              ' this same artifact
              ' for this demo, only 4 tile bitmaps defined

              ' empty tile

tile_square   word      %%3_3_3_3_3_3_3_3 ' tile 0
              word      %%3_0_0_0_0_0_0_3
              word      %%3_0_0_0_0_0_0_3
              word      %%3_0_0_0_0_0_0_3
              word      %%3_0_0_0_0_0_0_3
              word      %%3_0_0_0_0_0_0_3
              word      %%3_0_0_0_0_0_0_3
              word      %%3_3_3_3_3_3_3_3

Example: this particular bitmap definition is a square uses color 3 and fill with color 0,
Note: you may have up to 256 tiles


Format of palette array, each palette has 4 entries, each 1 byte, when a tile is defined in the tile map it has both an index
to the tile bitmap (0..255) as well as an index to the "palette" to use (0..255), below is an example of 4 palette entries

DAT

palette_map   ' 4 palettes for only now, palette memory (1..255) LONGs, each palette 4-BYTEs

              long $5C_5B_5A_02 ' palette 0
              long $0C_0B_0A_02 ' palette 1
              long $CC_CB_CA_02 ' paleete 2
              long $0C_0B_0A_02 ' palette 3

Note: each of the 8-bit values represents the color, luma, and chroma enable for the color as defined by the VSU hardware


}

HEL_GFX_Driver_Entry

              ' VCFG: setup Video Configuration register and 3-bit TV DAC pins to outputs

              mov       r0, par                         ' copy boot parameter value and read in parameters from main memory, must be on LONG boundary
              add       r0, #5*4
              rdlong  vcfg, r0
              movi    vcfg, #VIDEO_SETUP                ' vcfg I = controls overall setting, we want baseband video on bottom nibble, 2-bit color, enable chroma on broadcast & baseband

              add       r0, #4
              rdlong    r1, r0
              or      dira, r1                          ' set DAC pins to output 24, 25, 26 for hydra, 12, 13, 14 for Chameleon

              ' CTRA: setup Frequency to Drive Video
              movi    ctra, #VIDEO_CNTR_SETUP           ' pll internal routed to Video, PHSx+=FRQx (mode 1) + pll(16x)
                                                        ' needn't set D,S fields since they set pin A/B I/Os, but mode 1 is internal, thus irrelvant

              ' compute the value to place in FREQ such that the final counter
              ' output is NTSC and the PLL output is 16*NTSC
              mov     r1, v_freq                        ' r1 <- TV color burst frequency in Hz, eg. 3_579_545
              rdlong  r2, #CLKFREQ_REG                  ' r2 <- CLKFREQ is register 0, eg. 80_000_000
              call    #Dividefract                      ' perform r3 = 2^32 * r1 / r2

              mov     frqa, r3                          ' set frequency for counter such that bit 31 is toggling at a rate of the color burst (2x actually)
                                                        ' which means that the freq number added at a rate of CLKFREQ (usually 80.000 Mhz) results in a
                                                        ' pll output of the color burst, this is further multiplied by 16 as the final PLL output
                                                        ' thus giving the chroma hardware the clock rate of 16X color burst which is what we want :)


              mov       r0, par                         ' copy boot parameter value and read in parameters from main memory, must be on LONG boundary
              add       r0, #16
              mov       tile_status_bits_ptr, r0        ' ptr to status bits, so tile engine can pass out status of tile engine in real time


Next_Frame    ' start of new frame of 262 scanlines
              ' 26 top overscan
              ' 192 active vide
              ' 26 bottom overscan
              ' 18 vertical sync

              ' read run-time parameters from main memory, user can change these values every frame
              mov       r0, par                         ' copy boot parameter value and read in parameters from main memory, must be on LONG boundary
              rdlong    tile_map_base_ptr, r0           ' base ptr to tile map itself
              add       r0, #4
              rdlong    tile_bitmaps_base_ptr, r0       ' base pointer to array of 8x8 bitmaps, each 16 bytes
              add       r0, #4
              rdlong    tile_palettes_base_ptr, r0      ' base pointer to array of palettes, each palette 4 bytes / 1 long
              add       r0, #4
              rdlong    tile_map_control, r0            ' value of overscan color and tile map width 0=32, 1=64, 2=128, etc.

              mov       tile_map_overscan_color, tile_map_control     ' extract overscan color
              shr       tile_map_overscan_color, #16

              ' bug was OR'ing colors into overscan?? Which wasn't overwritten which it should be, FIXED
              mov        v_choverscan, tile_map_overscan_color        ' COPY into system overscan colors

              mov       tile_map_width, tile_map_control              ' extract tile map width
              and       tile_map_width, #$FF

              mov       r0, #0                          ' clear out status bits for next frame, 0 means region top overscan, vsync FALSE
              wrlong    r0, tile_status_bits_ptr        ' write out status bits to main memory


' /////////////////////////////////////////////////////////////////////////////
Top_Overscan_Scanlines

              mov       r0, #TSB_REGION_TOP_OVERSCAN    ' set region to top overscan
              wrlong    r0, tile_status_bits_ptr        ' write out status bits to main memory

              mov     r1, #26                           ' set # of scanlines

' Horizontal Scanline Loop (r1 itterations)
:Next_Scanline

              ' HSYNC 10.9us (Horizontal Sync) including color burst
              mov     vscl, v_shsync                    ' set the video scale register such that 16 serial pixels takes 39 color clocks (the total time of the hsync and burst)
              waitvid v_chsync, v_phsync                ' send out the pixels and the colors, the pixels are 0,1,2,3 that index into the colors, the colors are blank, sync, burst
                                                        ' we use them to create the hsync pulse itself

              ' set hsync output status bit and set region
              mov       r0, #TSB_REGION_TOP_OVERSCAN | TSB_HSYNC ' set region to top overscan and set hsync
              wrlong    r0, tile_status_bits_ptr        ' write out status bits to main memory


              ' HVIS 52.6us (Visible Scanline) draw 16 huge pixels
              mov     vscl, v_shvis                     ' set up the video scale so the entire visible scan is composed of 16 huge pixels
              waitvid v_choverscan , v_phoverscan_top   ' draw 16 pixels with overscan color

              mov       r0, #TSB_REGION_TOP_OVERSCAN    ' set region to top overscan and reset hsync
              wrlong    r0, tile_status_bits_ptr        ' write out status bits to main memory


              djnz    r1, #:Next_Scanline               ' are we done with the scanlines yet?

' /////////////////////////////////////////////////////////////////////////////
Active_Scanlines

              mov       r0, #TSB_REGION_ACTIVE_VIDEO    ' set region to active video
              wrlong    r0, tile_status_bits_ptr        ' write out status bits to main memory

              mov       r1, tile_map_control            ' reset scanline counter (consider scroll), in this case we count up
                                                        ' format $bc_tc_hv_ww
              shr       r1, #8                          ' shift vertical scroll value into lower bits
              and       r1, #$07                        ' mask off vertical scroll value and store
              mov       r7, r1                          ' r7 = r1+192, holds value of last line to render with scrolling
              add       r7, #192

              mov       tile_map_hscroll, tile_map_control ' now extract the horizontal scroll position
              shr       tile_map_hscroll, #12
              and       tile_map_hscroll, #$07             ' save for later

' Horizontal Scanline Loop (r1 iterations)
:Next_Scanline

              ' HSYNC 10.9us (Horizontal Sync) including color burst
              mov     vscl, v_shsync                    ' set the video scale register such that 16 serial pixels takes 39 color clocks (the total time of the hsync and burst)
              waitvid v_chsync, v_phsync                ' send out the pixels and the colors, the pixels are 0,1,2,3 that index into the colors, the colors are blank, sync, burst
                                                        ' we use them to create the hsync pulse itself

              ' set hsync output status bit and set region and active line
              mov       r0, r1
              shl       r0, #16                                     ' set the current line
              or        r0, #TSB_REGION_ACTIVE_VIDEO | TSB_HSYNC    ' set region to active video reset hsync
              wrlong    r0, tile_status_bits_ptr                    ' write out status bits to main memory

' /////////////////////////////////////////////////////////////////////////////

              ' this next section retrives the tile index, palette index for each tile and draws a single line of video composed of 1 sub
              ' scan line of each tile, the opportunity here is to realize that each tile accessed 16 times, once for each video line, thus
              ' the tile indexes and palettes themselves could be "cached" to save time, however, the added complexity is not needed yet
              ' but the next version will cache the data, so we can have more time per pixel block to do crazy stuff :)

              mov     vscl, v_spixel                    ' set up video scale register for 8 pixels at a time

              ' select proper sub-tile row address
              mov       r2, r1                          ' r1 holds current video line of active video (0..191)
              and       r2, #$07                        ' r2 = r1 mod 8, this is the sub-row in the tile we want to render, or the WORD index we need to offset the tile memory by

              ' access main memory can get the bitmap data
              shl       r2, #1                          ' r2 = r2*2, byte based row offset of pixels, 2 bytes per pixel row, 8 pixels, 2 bits per pixel
              add       r2, tile_bitmaps_base_ptr       ' r2 = tile_bitmaps_base_ptr + r2

              ' compute tile index itself for left edge of tile row, inner loop will index across as the scanline is rendered
              ' new code is needed here to support scrolling horizontally, we need to be able to tell the engine that a tile row has MORE than
              ' 8 tiles per row, so that when it access the next tile row down, it calculates properly, so this nice power of 2 math has to be
              ' revisited, however, we are going to constrain that the tiles per row must be a power of 2 multiple of 32 at least, so we don't have to use
              ' complex division yet, trying to keep each engine simple... so the playfield widths are  32, 64, 128, 256 (max), thus the
              ' only thing we need to add is a shift to the mask operation that performs the (line/8) * 32, since now the 32 could be 32, 64, 128, etc.

              ' new code with support for playfields 32, 64, 128... in size which then allows scrolling to be supported (course scrolling of course :)
              mov       tile_map_index_ptr, r1                ' r1 = line, copy it
              and       tile_map_index_ptr, #$1F8             ' tile_map_index_ptr = [(r1 / 8) * 32..256], this is the starting tile index for a row, 0, 32, 64, ...
              shl       tile_map_index_ptr, #2                ' results in tile_map_index_ptr = [(r1 / 8) * 32]
              shl       tile_map_index_ptr, tile_map_width    ' this finishes off the playfield x2, x4, x8, etc. scrolling support


              shl       tile_map_index_ptr, #1                ' tile_map_index_ptr = tile_index*2, since each tile is 2 bytes, we need to convert index to byte address
              add       tile_map_index_ptr, tile_map_base_ptr ' tile_map_index_ptr = [(r1 / 8) * 32..256] + tile_map_base_ptr, this is a byte address in main memory now


              ' at this point we have everything we need for the pixels rendering aspect of the 32 tiles that will be rendered, the inner loop logic will
              ' retrieve the time map indexes, and access the 8 pixels that make up each row of each tile, BUT we need to get the palette(s)
              ' for each tile as well, each tile has its own palette, but the palette will change each group of 8-pixels across the screen since
              ' each 8-pixels represents a single line from a different tile

              ' we could cache all the palettes into the local cache, but for fun let's just read them out of main memory during the inner loop

              ' render the 32 tile lines, r2 is holding proper row address, but we need to add base of actual tile we want rendered
              mov       r4, #32

:Pixel_Render_Loop

              ' read next tile index and palette index from main memory
              rdword    tile_map_word, tile_map_index_ptr

              ' retrieve 8-pixels of current row from proper bitmap referenced by tile
              mov       r3, tile_map_word
              and       r3, #$FF                        ' mask off upper 8-bits they hold the palette index, we aren't interested in
              shl       r3, #4                          ' r3 = tile_map_index*16 (bytes per tile bitmap)
              add       r3, r2                          ' r3 = tile_map_index*16 + tile_bitmaps_base_ptr + video_line mod 8
              rdword    r3, r3                          ' r3 = main_memory[r3], retrieve 16 bits of pixel data (8 pixels)
                                                        ' 16 clocks until hub comes around, try and be ready, move a couple instructions that aesthically
                                                        ' should be in one place between the hub reads to maximize processing/memory bandwith

              mov       v_pixels_buffer, r3             ' r3 holds pixels now, copy to pixel out buffer

              ' retrieve palette for current tile
              mov       r5, tile_map_word
              shr       r5, #8                          ' r5 now holds the palette index and we shifted out the tile index into oblivion
              shl       r5, #2                          ' multiple by 4, since there ar 4-bytes per palette entry
                                                        ' r5 = palette_map_index*4
              add       r5, tile_palettes_base_ptr      ' r5 = palette_map_base_ptr +palette_map_index*4

              ' moved from top of loop to eat time after previous memory read!
              add       tile_map_index_ptr, #2          ' advance pointer 2 bytes to next tile map index entry, for next pass

              rdlong    v_colors_buffer, r5             ' read the palette data into the buffer

              'draw the pixels with the selected palette (8 pixels, waitvid will come around 2x as fast now!)
              waitvid   v_colors_buffer, v_pixels_buffer

              ' reset hsync output status bit and set region and active line
              mov       r0, r1
              shl       r0, #16                         ' set the current line
              or        r0, #TSB_REGION_ACTIVE_VIDEO    ' set region to active video reset hsync
              wrlong    r0, tile_status_bits_ptr        ' write out status bits to main memory

              djnz      r4, #:Pixel_Render_Loop         ' loop until we draw 32 tiles (single pixel row of each)

' /////////////////////////////////////////////////////////////////////////////

              add       r1, #1
              cmp       r1, r7                  wc, wz  ' #192 normally
        if_b  jmp       #:Next_Scanline                 ' if ++r1 (current line) < 192 then loop

' /////////////////////////////////////////////////////////////////////////////

Bottom_Overscan_Scanlines

              mov       r0, #TSB_REGION_BOTT_OVERSCAN   ' set region to bottom overscan
              wrlong    r0, tile_status_bits_ptr        ' write out status bits to main memory

              mov       r1, #26                         ' set # of scanlines

' Horizontal Scanline Loop (r1 itterations)
:Next_Scanline


              ' HSYNC 10.9us (Horizontal Sync) including color burst
              mov       vscl, v_shsync                  ' set the video scale register such that 16 serial pixels takes 39 color clocks (the total time of the hsync and burst)
              waitvid   v_chsync, v_phsync              ' send out the pixels and the colors, the pixels are 0,1,2,3 that index into the colors, the colors are blank, sync, burst
                                                        ' we use them to create the hsync pulse itself
              ' set hsync output status bit and set region
              mov       r0, #TSB_REGION_BOTT_OVERSCAN | TSB_HSYNC ' set region to bottom overscan and set hsync
              wrlong    r0, tile_status_bits_ptr        ' write out status bits to main memory


              ' HVIS 52.6us (Visible Scanline) draw 16 huge pixels that represent an entire line of video
              mov       vscl, v_shvis                   ' set up the video scale so the entire visible scan is composed of 16 huge pixels
              waitvid   v_choverscan, v_phoverscan_bot  ' draw 16 pixels with bottom overscan color

              ' reset hsync output status bit and set region
              mov       r0, #TSB_REGION_BOTT_OVERSCAN   ' set region to bottom overscan and reset hsync
              wrlong    r0, tile_status_bits_ptr        ' write out status bits to main memory

              djnz      r1, #:Next_Scanline             ' are we done with the scanlines yet?

' /////////////////////////////////////////////////////////////////////////////
Vsync_Pulse

              ' VSYNC Pulse (Vertical Sync)
              ' 18 scanlines: 6 'high syncs', 6 'low syncs', and finally another 6 'high syncs'
              ' refer to NTSC spec, but this makes up the equalization pulses needed for a vsync

              mov       r0, #TSB_VSYNC | TSB_REGION_VSYNC ' set region to vsync as well
              wrlong    r0, tile_status_bits_ptr          ' write out vertical sync status bit to main memory

              call      #Vsync_High
              call      #Vsync_Low
              call      #Vsync_High
'
              jmp       #Next_Frame                     ' that's it, do it a googleplex times...

'//////////////////////////////////////////////////////////////////////////////
' SUB-ROUTINES VARIABLE DECLARATIONS
'//////////////////////////////////////////////////////////////////////////////

' /////////////////////////////////////////////////////////////////////////////
' vsync_high: Generate 'HIGH' vsync signal for 6 horizontal lines.
Vsync_High

              mov       r1, #6

              ' HSYNC 10.9us (Horizontal Sync)
:Vsync_Loop   mov       vscl, v_shsync
              waitvid   v_chsync, v_pvsync_high_1

              ' HVIS 52.6us (Visible Scanline)
              mov       vscl, v_shvis
              waitvid   v_chsync, v_pvsync_high_2
              djnz      r1, #:Vsync_Loop

Vsync_High_Ret
              ret

' /////////////////////////////////////////////////////////////////////////////
' vsync_low: Generate 'LOW' vsync signal for 6 horizontal lines.
Vsync_Low

              mov       r1, #6

              ' HSYNC 10.9us (Horizontal Sync)
:Vsync_Loop   mov       vscl, v_shsync
              waitvid   v_chsync, v_pvsync_low_1

              ' HVIS 52.6us (Visible Scanline)
              mov       vscl, v_shvis
              waitvid   v_chsync, v_pvsync_low_2
              djnz      r1, #:Vsync_Loop

Vsync_Low_Ret
              ret

' /////////////////////////////////////////////////////////////////////////////
' Calculates 2^32 * r1/r2, result stored in r3, r1 must be less that r2, that is, r1 < r2
' the results of the division are a binary weighted 32-bit fractional number where each bit
' is equal to the following weights:
' MSB (31)    30    29 ..... 0
'      1/2   1/4   1/8      1/2^32
Dividefract
              mov       r0,#32+1                        ' 32 iterations, are we done yet?
:Loop         cmpsub    r1,r2           wc              ' does divisor divide into dividend?
              rcl       r3,#1                           ' rotate carry into result
              shl       r1,#1                           ' shift dividend over
              djnz      r0,#:Loop                       ' done with division yet?

Dividefract_Ret
              ret                                       ' return to caller with result in r3

'//////////////////////////////////////////////////////////////////////////////
' VARIABLE DECLARATIONS
'//////////////////////////////////////////////////////////////////////////////

' general purpose registers

r0            long      $0
r1            long      $0
r2            long      $0
r3            long      $0
r4            long      $0
r5            long      $0
r6            long      $0
r7            long      $0

' tv output DAC port bit mask
'tvport_mask   long      %0000_0111 << 24        ' Hydra DAC is on bits 24, 25, 26
'tvport_mask   long      %0111_0000 << 8        ' Chameleon DAC is on bits 12, 13, 14


' output buffers to hold colors and pixels, start them off with "test" data
                        '3  2  1  0   <- color indexes
v_pixels_buffer long    %%1111222233330000
v_colors_buffer long    $5C_CC_0C_03            ' 3-RED | 2-GREEN | 1-BLUE | 0-BLACK

' pixel VSCL value for 256 visible pixels per line (clocks per pixel 8 bits | clocks per frame 12 bits )
' notice we divide PNTSC256 by 2, this in essence means draw only 8 pixels per waitvid, this way we can
' double the number of tiles, and only 8 of the 16 pixels in the "pixels" array sent to waitvid are rendered
v_spixel      long      ((PNTSC256 >> 4) << 12) + (PNTSC256 >> 1)


' hsync VSCL value (clocks per pixel 8 bits | clocks per frame 12 bits )
v_shsync      long      ((SNTSC >> 4) << 12) + SNTSC

' hsync colors (4, 8-bit values, each represent a color in the format chroma shift, chroma modulatation enable, luma | C3 C2 C1 C0 | M | L2 L1 L0 |
                        '3  2  1  0   <- color indexes
v_chsync      long      $00_00_02_8A ' SYNC (3) / SYNC (2) / BLACKER THAN BLACK (1) / COLOR BURST (0)

' hsync pixels
                        ' BP  |BURST|BW|    SYNC      |FP| <- Key BP = Back Porch, Burst = Color Burst, BW = Breezway, FP = Front Porch
'v_phsync      long      %%1_1_0_0_0_0_1_2_2_2_2_2_2_2_1_1
'v_phsync      long      %%1_1_1_1_0_0_0_0_1_2_2_2_2_1_1_1 ' perfect on all sets, looks like this is the winner!
v_phsync      long      %%1_1_1_1_0_0_0_1_2_2_2_2_2_1_1_1 ' perfect on all sets, looks like this is the winner!

' active video values
v_shvis       long      ((VNTSC >> 4) << 12) + VNTSC

' the colors used, 4 of them always
                        'red, color 3 | dark blue, color 2 | blue, color 1 | light blue, color 0
v_chvis       long      $5A_0A_0B_0C            ' each 2-bit pixel below references one of these 4 colors, (msb) 3,2,1,0 (lsb)

' the pixel pattern
v_phvis       long      %%3210_0123_3333_3333   ' 16-pixels, read low to high is rendered left to right, 2 bits per pixel
                                                ' the numbers 0,1,2,3 indicate the "colors" to use for the pixels, the colors
                                                ' are defined by a single byte each with represents the chroma shift, modulation,
                                                ' and luma
' the colors used, 4 of them always
                        'grey, color 3 | dark grey, color 2 | bottom overscan user controlled, color 1 | top overscan user controlled, color 0
v_choverscan  long      $07_04_00_00            ' each 2-bit pixel below references one of these 4 colors, (msb) 3,2,1,0 (lsb)

' the pixel pattern
v_phoverscan_top  long      %%0000_0000_0000_0000   ' 16-pixels, read low to high is rendered left to right, 2 bits per pixel
                                                    ' the numbers 0,1,2,3 indicate the "colors" to use for the pixels, the colors
                                                    ' are defined by a single byte each with represents the chroma shift, modulation,
                                                    ' and luma, always uses palette color 0

v_phoverscan_bot  long      %%1111_1111_1111_1111

' vsync pulses 6x High, 6x Low, 6x High
' the vertical sync pulse according to the NTSC spec should be composed of a series
' of pulses called the pre-equalization, serration pulses (the VSYNC pulse itself), and the post-equalization pulses
' there are 6 pulses of each, and they more or less inverted HSYNC, followed by 6 HSYNC pulses, followed by 6 more inverted HSYNC pulses.
' this keeps the horizontal timing circutry locked as well as allows the 60 Hz VSYNC filter to catch the "vsync" event.
' the values 1,2 index into "colors" that represent sync and blacker than black.
' so the definitions below help with generated the "high" and "low" dominate HSYNC timed pulses which are combined
' to generated the actual VSYNC pulse, refer to NTSC documentation for more details.

{
v_pvsync_high_1         long    %%1_1_1_1_1_1_1_1_1_1_1_2_2_2_1_1
v_pvsync_high_2         long    %%1_1_1_1_1_1_1_1_1_1_1_1_1_1_1_1

v_pvsync_low_1          long    %%2_2_2_2_2_2_2_2_2_2_2_2_2_2_1_1
v_pvsync_low_2          long    %%1_2_2_2_2_2_2_2_2_2_2_2_2_2_2_2
}

v_pvsync_high_1         long    %%1_1_1_1_1_1_1_1_1_1_2_2_2_2_1_1
v_pvsync_high_2         long    %%1_1_1_1_1_1_1_1_1_1_1_1_1_1_1_1

v_pvsync_low_1          long    %%2_2_2_2_2_2_2_2_2_2_2_2_2_1_1_1
v_pvsync_low_2          long    %%1_2_2_2_2_2_2_2_2_2_2_2_2_2_2_2


v_freq                  long    FNTSC

' tile engine locals
tile_row                long    $0
video_line              long    $0
tile_map_ptr            long    $0
tile_cache_ptr          long    $0
palette_cache_ptr       long    $0
tile_map_index          long    $0
tile_map_index_ptr      long    $0
tile_map_word           long    $0
tile_palette_index      long    $0

' tile engine passed parameters
tile_map_base_ptr       long    $0
tile_bitmaps_base_ptr   long    $0
tile_palettes_base_ptr  long    $0
tile_map_control        long    $0

' derived parms from passed parameters
tile_map_overscan_color long    $0
tile_map_width          long    $0

tile_map_vscroll        long    $0
tile_map_hscroll        long    $0

tile_status_bits_ptr    long    $0

' local COG cache memories (future expansion)
tile_cache              word    $0,$0,$0,$0,$0,$0,$0,$0,$0,$0,$0,$0,$0,$0,$0,$0
palette_cache           long    $0,$0,$0,$0,$0,$0,$0,$0,$0,$0,$0,$0,$0,$0,$0,$0

' END ASM /////////////////////////////////////////////////////////////////////
