/**
 * @file TvTileGfx.c - provides TV Tile Graphics via Andre's PASM Tile Graphics Driver
 *
 * Copyright (c) 2011, David Betz / Andre' LaMothe
 *
 * ----------------------------------------------------------------------------
 *                        TERMS OF USE: MIT License
 * ----------------------------------------------------------------------------
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 * ----------------------------------------------------------------------------
 */
 
#include "propeller.h"
#include "TvTileGfx.h"

/*
 * All functions are described in TvTileGfx.h
 */

int TvTileGfx_start(TvTileGfx_st *data, int basepin)
{
    extern uint32_t _binary_TvTileGfx_dat_start[];
    extern uint32_t _binary_TvTileGfx_dat_end[];
	int size = _binary_TvTileGfx_dat_end - _binary_TvTileGfx_dat_start;
	data->video_vcfg = ((basepin & 0x18) << 6) | ((basepin & 0x04) ? 0x70 : 0x07);
	data->video_dira = 0x7 << basepin;
    data->cogid = cognew(_binary_TvTileGfx_dat_start, size, (uint32_t *)PhysicalHubAddress(data)) + 1;
	return data->cogid;
}

void TvTileGfx_stop(TvTileGfx_st *data)
{
    if (data->cogid > 0)
        cogstop(--data->cogid);
}
