#include <stdio.h>
#include <stdint.h>
#include "propeller.h"
#include "dosfs.h"

#define SD_INIT_CMD             0x0d    // %011_01
#define SD_READ_CMD             0x11    // %100_01
#define SD_WRITE_CMD            0x15    // %101_01

static HUB uint32_t vm_params[3];

static void CopyAndSwapBytes(uint8_t *dst, uint8_t *src, int cnt);

int InitC3FileIO(int retries)
{
	int err;
	while (retries == 0 || --retries >= 0) {
		if ((err = _syscall(NULL, SYS_mboxcmd, PhysicalHubAddress(vm_mbox), SD_INIT_CMD)) == 0) 
			return InitFileIO();
		iprintf("Retrying SD init: %d\n", err);
	}
    return -1;
}

uint32_t DFS_ReadSector(uint8_t unit, uint8_t *buffer, uint32_t sector, uint32_t count)
{
    int err;
    while (count > 0) {
		vm_params[0] = PhysicalHubAddress(scratch_buffer);
        vm_params[1] = SECTOR_SIZE;
        vm_params[2] = sector;
        if ((err = _syscall(NULL, SYS_mboxcmd, PhysicalHubAddress(vm_mbox), SD_READ_CMD | (PhysicalHubAddress(vm_params) << 8))) != 0) {
			iprintf("SD_READ_CMD failed: %d\n", err);
            return -1;
		}
        CopyAndSwapBytes(buffer, scratch_buffer, SECTOR_SIZE);
        buffer += SECTOR_SIZE;
        ++sector;
        --count;
    }
    return 0;
}

uint32_t DFS_WriteSector(uint8_t unit, uint8_t *buffer, uint32_t sector, uint32_t count)
{
    int err;
    while (count > 0) {
        vm_params[0] = PhysicalHubAddress(scratch_buffer);
        vm_params[1] = SECTOR_SIZE;
        vm_params[2] = sector;
        CopyAndSwapBytes(scratch_buffer, buffer, SECTOR_SIZE);
        if ((err = _syscall(NULL, SYS_mboxcmd, PhysicalHubAddress(vm_mbox), SD_WRITE_CMD | (PhysicalHubAddress(vm_params) << 8))) != 0) {
			iprintf("SD_WRITE_CMD failed: %d\n", err);
            return -1;
		}
        buffer += SECTOR_SIZE;
        ++sector;
        --count;
    }
    return 0;
}

static void CopyAndSwapBytes(uint8_t *dst, uint8_t *src, int cnt)
{
    uint32_t *ldst = (uint32_t *)dst;
    uint32_t *lsrc = (uint32_t *)src;
    while ((cnt -= sizeof(uint32_t)) >= 0) {
		uint32_t value = *lsrc++;
		*ldst++ = (value >> 24)
               | ((value >>  8) & 0x0000ff00)
               | ((value <<  8) & 0x00ff0000)
               |  (value << 24);
	}
}
