#include <stdio.h>
#include "propeller.h"

#define DEL	0x7f

static int term_read(char *buf, int nbytes);
static int term_write(char *buf, int nbytes);

static int null_term_inbyte(void);
static void null_term_outbyte(int ch);
static int null_file_write(int fd, char *buf, int nbytes);
static int null_file_read(int fd, char *buf, int nbytes);

int (*term_inbyte_ptr)(void) = null_term_inbyte;
void (*term_outbyte_ptr)(int ch) = null_term_outbyte;
int (*file_read_ptr)(int fd, char *buf, int nbytes) = null_file_read;
int (*file_write_ptr)(int fd, char *buf, int nbytes) = null_file_write;

int read(int fd, char *buf, int nbytes)
{
    int cnt;
	switch (fd) {
	case 0:
		cnt = term_read(buf, nbytes);
		break;
	case 1:
	case 2:
		cnt = -1;
		break;
	default:
		cnt = (*file_read_ptr)(fd, buf, nbytes);
		break;
	}
	return cnt;
}

int write(int fd, char *buf, int nbytes)
{
    int cnt;
	switch (fd) {
	case 0:
		cnt = -1;
		break;
	case 1:
	case 2:
		cnt = term_write(buf, nbytes);
		break;
	default:
		cnt = (*file_write_ptr)(fd, buf, nbytes);
		break;
	}
	return cnt;
}

static int term_read(char *buf, int nbytes)
{
	int i = 0;
	while (i < nbytes) {
		int ch;
		if ((ch = (*term_inbyte_ptr)()) == -1)
			return -1;
		else if ((ch == '\r') || (ch == '\n')) {
			buf[i++] = '\n';
			(*term_outbyte_ptr)('\r');
			(*term_outbyte_ptr)('\n');
			break;
		}
		else if (ch == '\b' || ch == DEL) {
			if (i > 0) {
				(*term_outbyte_ptr)('\b');
				(*term_outbyte_ptr)(' ');
				(*term_outbyte_ptr)('\b');
				--i;
			}
		}
		else {
			buf[i++] = ch;
			(*term_outbyte_ptr)(ch);
		}
	}
	return i;
}

int inbyte(void)
{
	return (*term_inbyte_ptr)();
}

void outbyte(int ch)
{
	(*term_outbyte_ptr)(ch);
}

static int term_write(char *buf, int nbytes)
{
	int i = 0;
	while (i < nbytes)
		(*term_outbyte_ptr)(buf[i++]);
	return nbytes;
}

static int null_term_inbyte(void)
{
	return -1;
}

static void null_term_outbyte(int ch)  
{
}

static int null_file_write(int fd, char *buf, int nbytes)
{
	return nbytes;
}

static int null_file_read(int fd, char *buf, int nbytes)
{
	return -1;
}
