#include "propeller.h"

int errno;

HUB uint8_t scratch_buffer[2048];    // big enough for a COG image or a sector

/*
 * delay function used after start cog and during screen updates
 * @param per is number of ticks to wait ....
 */
void wait(int per)
{
    // not the best way to do it !
    volatile int time = 0;
    for(time = 0; time < per; )
        time++;
}    

void longfill(uint32_t *dst, uint32_t fill, int count)
{
	while (--count >= 0)
		*dst++ = fill;
}

void longmove(uint32_t *dst, uint32_t *src, int count)
{
	while (--count >= 0)
		*dst++ = *src++;
}

void longmove_swapwords(uint32_t *dst, uint32_t *src, int count)
{
	while (--count >= 0) {
		uint32_t dat = *src++;
		*dst++ = (dat << 16) | (dat >> 16);
	}
}

int cogid(void)
{
	return _syscall(&errno, SYS_cogid);
}

int cognew(uint32_t *code, int size, uint32_t *par)
{
    longmove((uint32_t *)scratch_buffer, code, size);
    return _syscall(&errno, SYS_coginit, 0x8, scratch_buffer, par);
}

int coginit(int cog, uint32_t *code, int size, uint32_t *par)
{
    longmove((uint32_t *)scratch_buffer, code, size);
	return _syscall(&errno, SYS_coginit, cog, scratch_buffer, par);
}

void cogstop(int cog)
{
	_syscall(&errno, SYS_cogstop, cog);
}
