#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include "propeller.h"

#define FILE_NAME		"test.dat"
#define FILE_NAME2		"test2.dat"

#define TEST_STRING		"abcdefghijklmnopqrstuvwxyz0123456789\n"
#define TEST_STRING2	"The quick brown fox jumps over the lazy dog.\n"

static void CreateFile(const char *name, const char *data, int len);
static int ReadFile(const char *name, char *data, int len);
static void Fatal(const char *msg);

int main(void)
{
    char buf[1024];
    int count;
    
	// write the first file
	CreateFile(FILE_NAME, TEST_STRING, sizeof(TEST_STRING) - 1);
	
	// read and verify the first file
	count = ReadFile(FILE_NAME, buf, sizeof(buf));
    iprintf("contents of %s: ", FILE_NAME);
    fwrite(buf, count, 1, stdout);
	if (strncmp(buf, TEST_STRING, strlen(TEST_STRING)) != 0)
		iprintf("compare failed\n");

	// write the second file
	CreateFile(FILE_NAME2, TEST_STRING2, sizeof(TEST_STRING2) - 1);
	
	// read and verify the second file
	count = ReadFile(FILE_NAME2, buf, sizeof(buf));
	iprintf("contents of %s: ", FILE_NAME2);
    fwrite(buf, count, 1, stdout);
	if (strncmp(buf, TEST_STRING2, strlen(TEST_STRING2)) != 0)
		iprintf("compare failed\n");

	// read and verify the first file again
	count = ReadFile(FILE_NAME, buf, sizeof(buf));
    iprintf("contents of %s: ", FILE_NAME);
    fwrite(buf, count, 1, stdout);
	if (strncmp(buf, TEST_STRING, strlen(TEST_STRING)) != 0)
		iprintf("compare failed\n");

    iprintf("Done\n");

    return 0;
}

static void CreateFile(const char *name, const char *data, int len)
{
	int fd, count, result;

    iprintf("Create file: %s\n", name);
    fd = open(name, O_WRONLY | O_CREAT);
    iprintf(" --> %d\n", fd);
	if (fd < 0)
		Fatal("file create failed");

    iprintf("Write file\n");
    count = write(fd, data, len);
    iprintf(" --> %d\n", count);
	if (count != len)
		Fatal("file write failed");

    iprintf("Close file\n");
    result = close(fd);
    iprintf(" --> %d\n", result);
	if (result != 0)
		Fatal("file close failed");
}

static int ReadFile(const char *name, char *data, int len)
{
	int fd, count, result;

    iprintf("Open file: %s\n", name);
    fd = open(name, O_RDONLY);
    iprintf(" --> %d\n", fd);
	if (fd < 0)
		Fatal("file open failed");

    iprintf("Read file\n");
    count = read(fd, data, len);
    iprintf(" --> %d\n", count);
	if (count < 0)
		Fatal("file read failed");

    iprintf("Close file\n");
    result = close(fd);
    iprintf(" --> %d\n", result);
	if (result != 0)
		Fatal("file close failed");

	return count;
}

static void Fatal(const char *msg)
{
	iprintf("error: %s\n", msg);
	exit(1);
}
