{
  Based on code from VMCOG - virtual memory server for the Propeller
  Copyright (c) February 3, 2010 by William Henning

  and on code from SdramCache.spin
  Copyright (c) 2010 by John Steven Denson (jazzed)

  and on code from PasmA0EEPROM.spin
  Copyright (c) 2009 by John Steven Denson (jazzed)

  TERMS OF USE: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
}

CON

  ' cache access commands
  WRITE_CMD             = %10
  READ_CMD              = %11

  ' extended commands
  ERASE_CHIP_CMD        = %000_01
  ERASE_BLOCK_CMD       = %001_01
  WRITE_DATA_CMD        = %010_01
  SD_INIT_CMD           = %011_01
  SD_READ_CMD           = %100_01
  SD_WRITE_CMD          = %101_01

  CMD_MASK              = %11
  EXTEND_MASK           = %10   ' this bit must be zero for an extended command

  LINELEN               = 1<<7  ' FIX ME!!!

  ' default cache dimensions
  DEFAULT_INDEX_WIDTH   = 5
  DEFAULT_OFFSET_WIDTH  = 7

  ' cache line tag flags
  EMPTY_BIT             = 30
  DIRTY_BIT             = 31

  ' address of CLKFREQ in hub RAM
  CLKFREQ_ADDR          = $0000

  'PLEN          = 32  ' standard EEPROM PAGE LENGTH
  PLEN          = 128  ' AT24C512 EEPROM PAGE LENGTH
  'PLEN          = 256 ' AT24C1024B EEPROM PAGE LENGTH

VAR
   long vm_mbox

PUB startx(mailbox, cache) | init_mbox, init_cache, init_iwidth, init_owidth
    vm_mbox := mailbox
    init_mbox := mailbox
    init_cache := cache
    init_iwidth := 5
    init_owidth := 7
    long[mailbox] := $ffffffff
    cognew(@init_vm, @init_mbox)
    repeat while long[mailbox]

pub readLong(madr)
    long[vm_mbox][0] := (madr&!3) | READ_CMD
    repeat while long[vm_mbox][0]
    madr &= LINELEN-1
    return long[long[vm_mbox][1]+madr]

pub writeLong(madr, value)
    long[vm_mbox][0] := (madr&!3) | WRITE_CMD
    repeat while long[vm_mbox][0]
    madr &= LINELEN-1
    long[long[vm_mbox][1]+madr] := value

pub readByte(madr)
    long[vm_mbox][0] := (madr&!3) | READ_CMD
    repeat while long[vm_mbox][0]
    madr &= LINELEN-1
    return byte[long[vm_mbox][1]+madr]

pub writeByte(madr, value)
    long[vm_mbox][0] := (madr&!3) | WRITE_CMD
    repeat while long[vm_mbox][0]
    madr &= LINELEN-1
    byte[long[vm_mbox][1]+madr] := value

pub eraseFlashBlock(madr)
    long[vm_mbox][0] := ERASE_BLOCK_CMD | (madr << 8)
    repeat while long[vm_mbox][0] <> 0
    return long[vm_mbox][1]

pub writeFlash(madr, buf, count_) | pbuf, pcnt, paddr
    pbuf := buf
    pcnt := count_
    paddr := madr
    long[vm_mbox][0] := WRITE_DATA_CMD | (@pbuf << 8)
    repeat while long[vm_mbox][0] <> 0
    return long[vm_mbox][1]

con
  EEID          = $A0
  RBIT          = 1   ' devid byte read bit
  WRBUF         = 1   ' if odd number of bits, write
  RDBUF         = 3   ' if even number of bits, read
  RDBYTES       = 4   ' if bit2 set, read small number of bytes

dat
  command      long  0          ' use for pins on startup
  address      long  0          ' read address
  count        long  0          ' bytes to read
  buffer       long  0          ' buffer
  earlyack     long  0          ' early ack for cache miss

dat ' just one singleton instance at a time please
  cog          long  0

        org   $0

' initialization structure offsets
' $0: pointer to a two word mailbox
' $4: pointer to where to store the cache lines in hub ram
' $8: number of bits in the cache line index if non-zero (default is DEFAULT_INDEX_WIDTH)
' $a: number of bits in the cache line offset if non-zero (default is DEFAULT_OFFSET_WIDTH)
' note that $4 must be at least 2^($8+$a)*2 bytes in size
' the cache line mask is returned in $0

init_vm mov     t1, par             ' get the address of the initialization structure
        rdlong  pvmcmd, t1          ' pvmcmd is a pointer to the virtual address and read/write bit
        mov     pvmaddr, pvmcmd     ' pvmaddr is a pointer into the cache line on return
        add     pvmaddr, #4
        add     t1, #4
        rdlong  cacheptr, t1        ' cacheptr is the base address in hub ram of the cache
        add     t1, #4
        rdlong  t2, t1 wz
  if_nz mov     index_width, t2     ' override the index_width default value
        add     t1, #4
        rdlong  t2, t1 wz
  if_nz mov     offset_width, t2    ' override the offset_width default value

        mov     index_count, #1
        shl     index_count, index_width
        mov     index_mask, index_count
        sub     index_mask, #1

        mov     line_size, #1
        shl     line_size, offset_width
        mov     t1, line_size
        sub     t1, #1
        wrlong  t1, par
        shr     line_size, #2

        ' set the pin directions
        mov     outa, spiout
        mov     dira, spidir

        ' select sequential access mode for the first SRAM chip
        call    #sram_chip1
        mov     data, ramseq
        mov     bits, #16
        call    #send

        ' select sequential access mode for the second SRAM chip
        call    #sram_chip2
        mov     data, ramseq
        mov     bits, #16
        call    #send

        call    #deselect

        ' unprotect the entire flash chip
        call    #write_enable
        call    #flash_chip
        mov     data, fwrstatus ' write zero to the status register
        mov     bits, #16
        call    #send
        call    #deselect

        ' get the clock frequency
        rdlong  sdFreq, #CLKFREQ_ADDR

        jmp     #vmflush

fillme  long    0[128-fillme]           ' first 128 cog locations are used for a direct mapped page table

        fit   128

        ' initialize the cache lines
vmflush movd    :flush, #0
        mov     t1, index_count
        shl     t1, #1                  ' clear both the flash and SRAM tags
:flush  mov     0-0, empty_mask
        add     :flush, dstinc
        djnz    t1, #:flush

        ' start the command loop
waitcmd mov     dira, #0                ' release the pins for other SPI clients
        wrlong  zero, pvmcmd
:wait   rdlong  vmpage, pvmcmd wz
  if_z  jmp     #:wait
        mov     dira, spidir            ' set the pins back so we can use them

        test    vmpage, #EXTEND_MASK wz ' test for an extended command
  if_z  jmp     #extend

        test    vmpage, bit20 wz        ' check for flash or SRAM access
        shr     vmpage, offset_width wc ' carry is now one for read and zero for write
        mov     set_dirty_bit, #0       ' make mask to set dirty bit on writes
        muxnc   set_dirty_bit, dirty_mask
        mov     line, vmpage            ' get the cache line index
        and     line, index_mask
  if_nz or      line, index_count       ' use upper entries for flash addresses
        mov     hubaddr, line
        shl     hubaddr, offset_width
        add     hubaddr, cacheptr       ' get the address of the cache line
        wrlong  hubaddr, pvmaddr        ' return the address of the cache line
        movs    :ld, line
        movd    :st, line
:ld     mov     vmcurrent, 0-0          ' get the cache line tag
        and     vmcurrent, tag_mask
        cmp     vmcurrent, vmpage wz    ' z set means there was a cache hit
  if_nz call    #miss                   ' handle a cache miss
:st     or      0-0, set_dirty_bit      ' set the dirty bit on writes
        jmp     #waitcmd                ' wait for a new command

' line is the cache line index
' vmcurrent is current page
' vmpage is new page
' hubaddr is the address of the cache line
miss    movd    :test, line
        movd    :st, line
:test   test    0-0, dirty_mask wz
  if_z  jmp     #:rd                    ' current page is clean, just read new page
        mov     vmaddr, vmcurrent
        shl     vmaddr, offset_width
        call    #BWRITE                 ' write current page
:rd     mov     vmaddr, vmpage
        shl     vmaddr, offset_width
        call    #BREAD                  ' read new page
:st     mov     0-0, vmpage
miss_ret ret

extend  mov     vmaddr, vmpage
        shr     vmaddr, #8
        shr     vmpage, #2
        and     vmpage, #7
        add     vmpage, #dispatch
        jmp     vmpage

dispatch
        jmp     #waitcmd
        jmp     #waitcmd
        jmp     #write_data_handler
        jmp     #waitcmd
        jmp     #waitcmd
        jmp     #waitcmd
        jmp     #waitcmd
        jmp     #waitcmd

write_data_handler
        jmp     #waitcmd


' pointers to mailbox entries
pvmcmd          long    0       ' on call this is the virtual address and read/write bit
pvmaddr         long    0       ' on return this is the address of the cache line containing the virtual address

cacheptr        long    0       ' address in hub ram where cache lines are stored
vmpage          long    0       ' page containing the virtual address
vmcurrent       long    0       ' current page in selected cache line (same as vmpage on a cache hit)
line            long    0       ' current cache line index
set_dirty_bit   long    0       ' DIRTY_BIT set on writes, clear on reads

zero            long    0       ' zero constant
dstinc          long    1<<9    ' increment for the destination field of an instruction
t1              long    0       ' temporary variable
t2              long    0       ' temporary variable

tag_mask        long    !(1<<DIRTY_BIT) ' includes EMPTY_BIT
index_width     long    DEFAULT_INDEX_WIDTH
index_mask      long    0
index_count     long    0
offset_width    long    DEFAULT_OFFSET_WIDTH
line_size       long    0                       ' line size in longs
empty_mask      long    (1<<EMPTY_BIT)
dirty_mask      long    (1<<DIRTY_BIT)

'----------------------------------------------------------------------------------------------------
'
' BREAD
'
' vmaddr is the virtual memory address to read
' hubaddr is the hub memory address to write
' count is the number of longs to read
'
' trashes fn, count, bits, data, ptr, count, c and z flags
'
'----------------------------------------------------------------------------------------------------

BREAD
BREAD_RET
        ret

'----------------------------------------------------------------------------------------------------
'
' BWRITE
'
' vmaddr is the virtual memory address to write
' hubaddr is the hub memory address to read
' count is the number of longs to write
'
' trashes fn, count, bits, data, ptr, count, z flag
'
'----------------------------------------------------------------------------------------------------

BWRITE
BWRITE_RET
        ret

vmaddr      long    0       ' virtual address
hubaddr     long    0       ' hub memory address to read from or write to

' temporaries used by BREAD and BWRITE
ptr         long    0
count       long    0

' temporaries used by send
bits        long    0
data        long    0

read        long    $03000000       ' read command
write       long    $02000000       ' write command
ramseq      long    $01400000       ' %00000001_01000000 << 16 ' set sequential mode
readstat    long    $05000000       ' read status

scl            long 1<<28   ' <-- res 1
sda            long 1<<29

tp             long 0   ' temp pointer
ap             long 0   ' addr pointer
bp             long 0   ' buff pointer
sp             long 0   ' size/count pointer

t0             long 0   ' temp vars
't1             long 0

cmd            long 0   ' command word
devs           long 0   ' device select
addr           long 0   ' address
bytes          long 0   ' size in bytes
psiz           long 0   ' parameter size in bytes
buff           long 0   ' buffer pointer
ueack          long 0   ' user early ack
ack            long 0   ' ack bit
val            long 0   ' accumulated value

'zero           long 0

'-------------------------------------------------------------------------------
' start point
'
init

  mov          tp,  par ' long 0 word 0
  rdword       t0,  tp
  mov          scl, #1
  shl          scl, t0  ' scl pin bit    

  add          tp, #2   ' word 1
  rdword       t0,  tp
  mov          sda, #1
  shl          sda, t0  ' sda pin bit
    
  add          tp,  #2  ' long 1
  mov          ap,  tp  ' address pointer
  add          tp,  #4  ' long 3
  mov          sp,  tp  ' size/count pointer
  add          tp,  #4  ' long 4
  mov          bp,  tp  ' buffer pointer
  add          tp,  #4  ' long 5
  mov          ueack,tp ' early ack pointer

:initlp0
  mov          outa,    scl             ' scl 1
  mov          dira,    scl             ' scl output
  andn         dira,    sda             ' sda output

  jmp          #cmdDone ' clear par so we don't have command pending at startup

'-------------------------------------------------------------------------------
' command loop
'
cmdDone
  wrlong       zero,    par
doCmd
  rdlong       cmd,     par wz
  mov          devs,    #EEID           ' get device select ... always EEID
  if_z  jmp    #doCmd

  rdlong       bytes,   sp              ' get number of bytes to read
  rdlong       addr,    ap              ' get address to read
  rdlong       tp,      bp              ' get buffer address for read


  cmp          cmd,     #RDBYTES wc,wz  ' if bit 2 set, use early ack
  if_e  jmp    #cmdReadBytes

  test         cmd,     #RDBUF wc       ' if c, odd bits for write
  if_nc jmp    #cmdReadBuf              ' else even bits for read
  jmp          #cmdWriteBuf

  jmp          #cmdDone                 ' this should never run
'
'-------------------------------------------------------------------------------


{{
'-------------------------------------------------------------------------------
' read buffer from eeprom
' @param       devs - device number
' @param       addr - read address
' @param       bytes - number of bytes to read
' @param       tp   - address of buffer for read
'
}}
cmdReadBuf
:loop
  cmp          bytes,   #PLEN wc,wz
  if_be mov    psiz,    bytes
  if_a  mov    psiz,    #PLEN
  call         #readPage
  add          addr,    psiz
  sub          bytes,   psiz wc,wz
  if_a  jmp    #:loop
  jmp          #cmdDone

{{
'-------------------------------------------------------------------------------
' write buffer to eeprom
' @param       *ap - write address
' @param       *sp - number of bytes to write
' @param       *bp - address of buffer to write
'
}}
cmdWriteBuf
:loop
  cmp          bytes,   #PLEN wc,wz
  if_be mov    psiz,    bytes
  if_a  mov    psiz,    #PLEN
  call         #writePage
  add          addr,    psiz
  sub          bytes,   psiz wc,wz
  if_a  jmp    #:loop
  jmp          #cmdDone


'-------------------------------------------------------------------------------
' read eeprom page
' @param       *dp - device number
' @param       *ap - read address
' @param       *bp - address of buffer for read
' @param       psiz - number of bytes to read
'
readPage
  call         #sendAddr
  or           devs,    #RBIT   ' send read command
  call         #sendCurrent
  mov          t1,      psiz    ' read psiz bytes
  sub          t1,      #1      ' first psiz-1 are NAK
  tjz          t1,      #:last
  andn         ack,     sda     ' not last byte     
:loop
  call         #recVal          ' read page data
  wrbyte       val,     tp
  add          tp,      #1
  djnz         t1,      #:loop
:last  
  or           ack,     sda     ' last byte has ACK     
  call         #recVal
  wrbyte       val,     tp
  add          tp,      #1      ' increment for next loop too
  call         #sendEnd
readPage_ret   ret

{{
'-------------------------------------------------------------------------------
' read buffer from eeprom with early ack after 1 byte read
' @param       devs - device number
' @param       addr - read address
' @param       bytes - number of bytes to read
' @param       tp   - address of buffer for read
'
}}
cmdReadBytes
  'mov          psiz,    bytes
  'call         #readPage
  call         #sendAddr
  or           devs,    #RBIT   ' send read command
  call         #sendCurrent
  mov          t1,      bytes   ' read psiz bytes
  'sub          t1,      #1      ' first psiz-1 are NAK
  tjz          t1,      #:last
  andn         ack,     sda     ' not last byte
:loop
  call         #recVal          ' read page data
  wrbyte       val,     tp
  add          tp,      #1
  djnz         t1,      #:loop
:last
  or           ack,     sda     ' last byte has ACK
  call         #recVal
  wrbyte       val,     tp
  add          tp,      #1      ' increment for next loop too
  call         #sendEnd
  jmp          #cmdDone

'-------------------------------------------------------------------------------
' read current byte
' @param       devs - device number
' @param       addr - read address
' @param       bytes - number of bytes to read
' @param       tp   - address of buffer for read
'
readCurrent
  or           devs,    #RBIT   ' read
  call         #sendCurrent
  or           ack,     sda     ' last byte     
  call         #recVal
  wrbyte       val,     tp
  add          tp,      #1
  call         #sendEnd
readCurrent_ret   ret

'-------------------------------------------------------------------------------
' write eeprom page
' @param       devs - device number
' @param       addr - read address
' @param       bytes - number of bytes to read
' @param       tp   - address of buffer for read
' @param       psiz - number of bytes to read
'
writePage
  call         #sendAddr        ' send page address
  mov          t1,      psiz
:loop
  rdbyte       val,     tp
  add          tp,      #1
  call         #sendVal         ' send data
  djnz         t1,      #:loop
  call         #writeEndPoll
writePage_ret  ret

'-------------------------------------------------------------------------------
' send end and poll for write complete
'
writeEndPoll
  call         #sendEnd         ' send stop
  call         #sclDelay
  call         #sendCurrent     ' now start write polling
  or           outa,    scl
  andn         dira,    sda     ' looking for ACK
:loop
  test         sda,     ina wz  ' is ack 0 ?
  if_nz jmp    #:loop           ' if ack 1, repeat
writeEndPoll_ret  ret  

'-------------------------------------------------------------------------------
' write current eeprom byte
' @param       devs - device number
' @param       addr - read address
' @param       bytes - number of bytes to read
' @param       tp   - address of buffer for read
'
writeCurrent
  andn         devs,    #RBIT
  call         #sendStart
  rdbyte       val,     tp
  add          tp,      #1
  call         #sendVal
writeCurrent_ret  ret

'-------------------------------------------------------------------------------
' send eeprom address
' @param       devs - device number
' @param       addr - read address

'
sendAddr
  call         #sendStart
  mov          val,     addr            ' %00000000_00000111_11111111_11111111 ->
  ror          val,     #16             ' %11111111_11111111_00000000_00000111
  shl          val,     #1              ' %11111111_11111110_00000000_00001110 ->
  andn         devs,    #RBIT           ' %10100000 send addr/write needs bit 0 low 
  or           val,     devs            ' %11111111_11111110_00000000_10101110
  call         #sendVal                 ' send device id and upper address bits
  mov          val,     addr            ' %00000000_00000111_11111111_11111111 ->
  ror          val,     #8              ' %11111111_00000000_00000111_11111111
  call         #sendVal                 ' send middle address bits
  mov          val,     addr            ' %00000000_00000111_11111111_11111111
  call         #sendVal                 ' send lower address bits
sendAddr_ret   ret

'-------------------------------------------------------------------------------
' send current is used to initiate a transaction after sendAddr
' @param       devs - device number with bit 0 set for read or clear for write
' @param       addr - read address
'
sendCurrent
  call         #sendStart
  mov          val,     addr            ' %00000000_00000111_11111111_11111111 ->
  ror          val,     #16             ' %11111111_11111111_00000000_00000111
  shl          val,     #1              ' %11111111_11111110_00000000_00001110 ->
  or           val,     devs            ' %11111111_11111110_00000000_1010111r
  call         #sendVal                 ' send device id and upper address bits
sendCurrent_ret  ret


'-------------------------------------------------------------------------------
' A start signal is when SDA goes 1 -> 0 while SCL is 1
'
sendStart
  or           outa,    scl             ' scl high
  or           dira,    scl             ' scl output ... but it is always output right?
  or           outa,    sda             ' sda high
  or           dira,    sda             ' sda output
  call         #sclDelayD2              ' device detects high
  andn         outa,    sda             ' sda low for start condition 
  call         #sclDelayD2              ' device detects start
  andn         outa,    scl             ' finish with scl low 
sendStart_ret  ret
  
'-------------------------------------------------------------------------------
' An end signal is when SDA goes 0 -> 1 while SCL is 1
'
sendEnd
  andn         outa,    sda             ' sda high
  or           outa,    scl             ' scl high
  call         #sclDelayD2
  or           outa,    sda             ' sda high
  call         #sclDelayD2              ' device detects stop
  andn         dira,    sda             
  andn         dira,    scl             ' leave sda & scl input so other devices can use the bus
sendEnd_ret    ret

'-------------------------------------------------------------------------------
' scl has limits on how fast it can go. use this routine to throttle scl
'
sclDelay
  call         #sclDelayD2
  call         #sclDelayD2
sclDelay_ret   ret

sclDelayD2
  nop
  nop
sclDelayD2_ret ret

'-------------------------------------------------------------------------------
' send 8 bits of data
' @param ack
' @param val
'
sendVal
  mov          ack,     #0              ' %00000000_00000000_00000000_11111111 ->
  ror          val,     #8              ' %11111111_00000000_00000000_00000000
  andn         outa,    scl             ' start with scl 0
  mov          t0,      #8              ' Output val on sda
:loop
  shl          val,     #1 wc           ' get next bit state into carry              
  muxc         outa,    sda             ' set bit depending on carry
  call         #sclDelay
  or           outa,    scl             ' toggle clock
  call         #sclDelay
  andn         outa,    scl
  call         #sclDelayD2
  djnz         t0,      #:loop
:findack  
  andn         dira,    sda             ' looking for ACK
  mov          ack,     ina             ' get ack
  and          ack,     sda wz          ' keep only ack bit
  or           outa,    scl             ' scl high
  call         #sclDelay
  andn         outa,    scl             ' scl low
  andn         outa,    sda             ' sda low
  or           dira,    sda             ' sda back to output
sendVal_ret    ret


'-------------------------------------------------------------------------------
' receive 8 bits of data
' @param ack
' @param val
'
recVal
  mov          val,     #0
  andn         dira,    sda             ' sda input
  andn         outa,    scl             ' clock starts at 0
  call         #sclDelay
  mov          t0,      #8              ' input 8 bits of val from sda
:loop
  or           outa,    scl             ' toggle clock
  call         #sclDelayD2
  shl          val,     #1
  test         sda,     ina wz          ' get bit in z
  muxnz        val,     #1
  andn         outa,    scl             ' toggle clock
  call         #sclDelay
  djnz         t0,      #:loop
:sendack
  or           dira,    sda             ' set sda to out so device can interpret ACK
  or           outa,    ack             ' ack is always on sda, but not always set
  call         #sclDelayD2
  or           outa,    scl        
  call         #sclDelay
  andn         outa,    scl             ' final clock pulse
  andn         outa,    sda        
recVal_ret     ret

'pad            long 0 [($1fc-pad)]      ' pad out to 2KB ... the HUB memory startup area
                                         ' provides a buffer for loading COGS from eeprom

        FIT     496             ' out of 496

{{
 TERMS OF USE: MIT License                                                           

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
}}
