' this selects the mailbox interface
'#define USE_JCACHED_MEMORY

' select only one of the below
'#define USE_SDRAM_MEMORY        ' John Steven Denson (jazzed)'s SDRAM support
'#define USE_C3_SPI_MEMORY       ' Parallax's C3 SPI flash/SRAM support

' select debugging output to a TV
#define TV_DEBUG

#ifdef USE_SDRAM_MEMORY
#define XCACHE
#endif

CON
  _clkmode = def#_clkmode
  _xinfreq = def#_xinfreq

  ' character codes
  CR  = $0d

  hub_memory_size = 32 * 1024
  rendezvous_area_size = 128
  zog_dispatch_table_size = 128

#ifdef USE_C3_SPI_MEMORY
  zpu_memory_size = 64 * 1024
#endif
#ifdef USE_SDRAM_MEMORY
  zpu_memory_size = 32 * 1024 * 1024
#endif
  vm_cache_size = 8 * 1024
  vm_mbox_size = 2 * 4
  zog_mbox_size = 8 * 4
  debug_control_size = 10 * 4
  debug_buffer_size = 16 * 2
  rendezvous_area = hub_memory_size - rendezvous_area_size
  zog_dispatch_table = rendezvous_area - zog_dispatch_table_size
  vm_cache = zog_dispatch_table - vm_cache_size
  vm_mbox = rendezvous_area
  zog_mbox = vm_mbox + vm_mbox_size
  debug_control = zog_mbox + zog_mbox_size
  debug_buffer = debug_control + debug_control_size

OBJ
#ifdef USE_C3_SPI_MEMORY
  def  : "userdefs_c3.spin"
  vm   : "spisram_cache_c3"
#endif
#ifdef USE_SDRAM_MEMORY
  def  : "userdefs_sdram.spin"
  vm   : "SdramCache"
#endif
  zog  : "zog.spin"
#ifdef TV_DEBUG
  tv   : "TV_Text"
#endif

VAR
  '8 consecutive longs for Zog's par block.
  long par_zpu_memory_addr
  long par_zpu_memory_sz
  long par_initial_pc
  long par_initial_sp
  long par_dispatch_tab_addr
  long par_pasm_addr
  long par_zog_mbox_addr
  long par_vm_mbox_addr

PUB start

#ifdef TV_DEBUG
  tv.start(def#TvPin)
  tv.str(string("SD Flash Loader v0.1 ("))
  tv.str(@board_name)
  tv.str(string(")", CR))
#endif

  vm.startx(vm_mbox, vm_cache)
  start_bytecode_in_flash
  start_zog

#ifdef TV_DEBUG

PRI crlf
  tv.out(CR)

PRI taghex8(tag, addr)
  tv.str(tag)
  tv.hex(addr, 8)
  crlf

#endif

PRI swap(value) | swapped
  return value
  swapped :=  value              << 24
  swapped |= (value & $0000ff00) << 8
  swapped |= (value & $00ff0000) >> 8
  swapped |=  value              >> 24
  return swapped

PUB start_bytecode_in_flash | data_image, data_start, data_end, bss_start, bss_end

  ' load the image header
  par_initial_pc := swap(vm.readLong($00100000))
  data_image     := swap(vm.readLong($00100004))
  data_start     := swap(vm.readLong($00100008))
  data_end       := swap(vm.readLong($0010000c))
  bss_start      := swap(vm.readLong($00100010))
  bss_end        := swap(vm.readLong($00100014))

#ifdef TV_DEBUG
  taghex8(string("start:      "), par_initial_pc)
  taghex8(string("data_image: "), data_image)
  taghex8(string("data_start: "), data_start)
  taghex8(string("data_end:   "), data_end)
  taghex8(string("bss_start:  "), bss_start)
  taghex8(string("bss_end:    "), bss_end)
#endif

  ' initialize .data
  repeat while data_start < data_end
    vm.writeLong(data_start, vm.readLong(data_image))
    data_image += 4
    data_start += 4

  ' clear .bss
  repeat while bss_start < bss_end
    vm.writeLong(bss_start, 0)
    bss_start += 4

PUB start_zog | cog

  'Set up Zog's parameter block (except par_initial_pc)
  par_zpu_memory_addr := 0      ' using external memory
  par_zpu_memory_sz := 0        ' using external memory
  par_initial_sp := zpu_memory_size - 8
  par_dispatch_tab_addr := zog.getdispatch_table
  par_pasm_addr := zog.getzog
  par_zog_mbox_addr := zog_mbox
  par_vm_mbox_addr := vm_mbox

  waitcnt(cnt+clkfreq)                               'Give serial terminal window time

#ifdef TV_DEBUG
  tv.stop
#endif

  ' let ZOG replace the Spin loader code
  coginit(cogid, @boot_zog, @par_zpu_memory_addr)

PRI CheckError(r)
  if r < 0
#ifdef TV_DEBUG
    tv.str(string("Error: "))
    tv.hex(r,2)
    tv.str(string(" HALTED."))
    crlf
#endif
    repeat                                             '<== LOOP HERE

DAT
                        org     0
'Copy PAR block into COG
boot_zog                mov     zog_params, par
                        mov     address, zog_params
                        add     address, #16
                        rdlong  dispatch_tab_addr, address
                        wrlong  dispatch_tab_final, address 'Update param block with new address
                        add     address, #4
                        rdlong  pasm_addr, address

'Move ZOG dispatch table runtime position in HUB RAM
move_dispatch           mov     src, dispatch_tab_addr
                        mov     dest, dispatch_tab_final
                        mov     count, #$80
:loop                   rdbyte  temp, src
                        wrbyte  temp, dest
                        add     src, #1
                        add     dest, #1
                        djnz    count, #:loop

'Launch Zog in this COG.
go_zog                  cogid   coginit_dest            'Get ID of this COG for restarting
                        mov     temp, pasm_addr         'Set ZOG PASM address for coginit
                        shl     temp, #2                'Shift into bits 17:4 (only high 14 bits required)
                        or      coginit_dest, temp      'Place in dest for coginit

                        mov     temp, zog_params        'Set PAR block address for coginit
                        shl     temp, #16               'Move to bits 31:18 (only high 14 bits required)
                        or      coginit_dest, temp      'Combine PASM addr and PAR addr

                        coginit coginit_dest

#ifdef USE_C3_SPI_MEMORY
board_name              byte    "C3", 0
#endif
#ifdef USE_SDRAM_MEMORY
board_name              byte    "SDRAM", 0
#endif

zog_params              long    0
dispatch_tab_addr       long    0
pasm_addr               long    0
address                 long    0
dest                    long    0
src                     long    0
count                   long    0
temp                    long    0
coginit_dest            long    0

'The final run time position of Zog's dispatch table
dispatch_tab_final      long    zog_dispatch_table

                        fit     $01ff

'The End.

