' this selects the mailbox interface
'#define USE_JCACHED_MEMORY

' select only one of the below
'#define USE_SDRAM_MEMORY        ' John Steven Denson (jazzed)'s SDRAM support
'#define USE_C3_SPI_MEMORY       ' Parallax's C3 SPI flash/SRAM support

' select debugging output to a TV
#define TV_DEBUG

#ifdef USE_SDRAM_MEMORY
#define XCACHE
#endif

CON
  _clkmode = def#_clkmode
  _xinfreq = def#_xinfreq

  ' character codes
  CR  = $0d

  hub_memory_size = 32 * 1024
  rendezvous_area_size = 128
  zog_dispatch_table_size = 128

#ifdef USE_C3_SPI_MEMORY
  zpu_memory_size = 64 * 1024
#endif
#ifdef USE_SDRAM_MEMORY
  zpu_memory_size = 32 * 1024 * 1024
#endif
  vm_cache_size = 8 * 1024
  vm_mbox_size = 2 * 4
  zog_mbox_size = 8 * 4
  debug_control_size = 10 * 4
  debug_buffer_size = 16 * 2
  rendezvous_area = hub_memory_size - rendezvous_area_size
  zog_dispatch_table = rendezvous_area - zog_dispatch_table_size
  vm_cache = zog_dispatch_table - vm_cache_size
  vm_mbox = rendezvous_area
  zog_mbox = vm_mbox + vm_mbox_size
  debug_control = zog_mbox + zog_mbox_size
  debug_buffer = debug_control + debug_control_size

OBJ
#ifdef USE_C3_SPI_MEMORY
  def  : "userdefs_c3.spin"
  vm   : "spisram_cache_c3"
#endif
#ifdef USE_SDRAM_MEMORY
  def  : "userdefs_sdram.spin"
  vm   : "SdramCache"
#endif
  zog  : "zog.spin"
  sd   : "fsrwFemto_rr001"
#ifdef TV_DEBUG
  tv   : "TV_Text"
#endif

VAR
  '8 consecutive longs for Zog's par block.
  long par_zpu_memory_addr
  long par_zpu_memory_sz
  long par_initial_pc
  long par_initial_sp
  long par_dispatch_tab_addr
  long par_pasm_addr
  long par_zog_mbox_addr
  long par_vm_mbox_addr

  long ioControl[2]   ' SD parameters
  byte diskbuff[512]  ' SD sector buffer
  byte filename[129]  ' Image filename
  long start_ram_image

PUB start

#ifdef TV_DEBUG
  tv.start(def#TvPin)
  tv.str(string("SD Loader v0.2 ("))
  tv.str(@board_name)
  tv.str(string(")", CR))
#endif

  start_ram_image := 0

  vm.startx(vm_mbox, vm_cache)
  mountSD
  read_zogload_run
  load_image
  if start_ram_image
    start_bytecode_in_ram
  else
#ifdef USE_C3_SPI_MEMORY
    start_bytecode_in_flash
#endif
#ifdef USE_SDRAM_MEMORY
    Fatal(string("Flash memory not supported"))
#endif
  start_zog

#ifdef TV_DEBUG

PRI crlf
  tv.out(CR)

PRI taghex8(tag, addr)
  tv.str(tag)
  tv.hex(addr, 8)
  crlf

#endif

PRI mountSD | err
  if sd.startSD(@ioControl) == 0
    return 0
  repeat
#ifdef TV_DEBUG
    tv.str(string("Mounting SD card...", CR))
#endif
    err := \sd.mount_explicit(def#spiDO,def#spiClk,def#spiDI,def#spiCS)
  until err == 0
  return 1

PUB read_zogload_run | bytes, err
#ifdef TV_DEBUG
  tv.str(string("Opening 'zogload.run'..."))
#endif
  err := \sd.popen(string("zogload.run"), "r")
#ifdef TV_DEBUG
  tv.hex(err, 8)
  crlf
  tv.str(string("Reading filename..."))
#endif
  bytes := sd.pread(@filename, 128)
  CheckError(bytes)
  byte[@filename][bytes] := 0
#ifdef TV_DEBUG
  tv.str(@filename)
  crlf
#endif
  err := sd.pclose
  CheckError(err)

PUB load_image | bytes, err
#ifdef TV_DEBUG
  tv.str(string("Opening ZPU image: "))
  tv.str(@filename)
  tv.str(string("..."))
#endif
  err := \sd.popen(@filename, "r")
#ifdef TV_DEBUG
  tv.hex(err, 8)
  crlf
  tv.str(string("Reading image..."))
#endif

  bytes := sd.pread(@diskbuff, 512)
  if bytes < 4
    Fatal(string("Bad image header"))

  if long[@diskbuff] == $00000000
    load_bytecode(bytes)
  else
#ifdef USE_C3_SPI_MEMORY
    write_bytecode_to_flash(bytes)
#endif
#ifdef USE_SDRAM_MEMORY
    Fatal(string("Flash memory not supported"))
#endif

PUB load_bytecode(bytes) | addr, val, i, err
  addr := 0
  err := sd.stopSDCard                                'Stop SD card, but do not stop cog
  CheckError(err)
  repeat while bytes > 0
    repeat i from 0 to bytes - 1 step 4
      val := (diskbuff[i] << 24) | (diskbuff[i+1] << 16) | (diskbuff[i+2] << 8) | diskbuff[i+3]
      vm.writeLong(addr, val)
      addr += 4
    err := \sd.initSDCard(def#spiDO,def#spiClk,def#spiDI,def#spiCS)     'Init the SD
    bytes := sd.pread(@diskbuff, 512)
    err := sd.stopSDCard                                'Stop SD card, but do not stop cog
    CheckError(err)
#ifdef TV_DEBUG
  tv.dec(addr)
  tv.str(string(" bytes"))
  crlf
#endif
  start_ram_image := 1

PUB start_bytecode_in_ram | load_address, bss_start, bss_end, i

  ' load the image header
  load_address   := swap(vm.readLong($00000000))
  par_initial_pc := swap(vm.readLong($00000004))
  bss_start      := swap(vm.readLong($00000008))
  bss_end        := swap(vm.readLong($0000000c))

#ifdef TV_DEBUG
  taghex8(string("load_address: "), load_address)
  taghex8(string("start:        "), par_initial_pc)
  taghex8(string("bss_start:    "), bss_start)
  taghex8(string("bss_end:      "), bss_end)
#endif

  ' clear .bss
  repeat while bss_start < bss_end
    vm.writeLong(bss_start, 0)
    bss_start += 4

#ifdef USE_C3_SPI_MEMORY

PUB write_bytecode_to_flash(bytes) | i, addr, err
  addr := 0
  err := sd.stopSDCard                                'Stop SD card, but do not stop cog
  CheckError(err)
  vm.eraseFlashBlock(addr)
  repeat while bytes > 0
    if bytes > 0
      vm.writeFlash(addr, @diskbuff, bytes)
      addr += bytes
      if (addr & $00000fff) == 0
        vm.eraseFlashBlock(addr)
      err := \sd.initSDCard(def#spiDO,def#spiClk,def#spiDI,def#spiCS)     'Init the SD
      bytes := sd.pread(@diskbuff, 512)
      err := sd.stopSDCard                                'Stop SD card, but do not stop cog
      CheckError(err)
#ifdef TV_DEBUG
  tv.dec(addr)
  tv.str(string(" bytes"))
  crlf
#endif
  start_ram_image := 0

PUB start_bytecode_in_flash | load_address, bss_start, bss_end, data_image, data_start, data_end

  ' load the image header
  load_address   := swap(vm.readLong($00100000))
  par_initial_pc := swap(vm.readLong($00100004))
  bss_start      := swap(vm.readLong($00100008))
  bss_end        := swap(vm.readLong($0010000c))
  data_image     := swap(vm.readLong($00100010))
  data_start     := swap(vm.readLong($00100014))
  data_end       := swap(vm.readLong($00100018))

#ifdef TV_DEBUG
  taghex8(string("load_address: "), load_address)
  taghex8(string("start:        "), par_initial_pc)
  taghex8(string("bss_start:    "), bss_start)
  taghex8(string("bss_end:      "), bss_end)
  taghex8(string("data_image:   "), data_image)
  taghex8(string("data_start:   "), data_start)
  taghex8(string("data_end:     "), data_end)
#endif

  ' initialize .data
  repeat while data_start < data_end
    vm.writeLong(data_start, vm.readLong(data_image))
    data_image += 4
    data_start += 4

  ' clear .bss
  repeat while bss_start < bss_end
    vm.writeLong(bss_start, 0)
    bss_start += 4

#endif

PUB start_zog | cog

  'Set up Zog's parameter block (except par_initial_pc)
  par_zpu_memory_addr := 0      ' using external memory
  par_zpu_memory_sz := 0        ' using external memory
  par_initial_sp := zpu_memory_size - 8
  par_dispatch_tab_addr := zog.getdispatch_table
  par_pasm_addr := zog.getzog
  par_zog_mbox_addr := zog_mbox
  par_vm_mbox_addr := vm_mbox

  waitcnt(cnt+clkfreq)                               'Give serial terminal window time

  ' stop all COGs except the one running the vm
  sd.stop

#ifdef TV_DEBUG
  tv.stop
#endif

  ' let ZOG replace the Spin loader code
  coginit(cogid, @boot_zog, @par_zpu_memory_addr)

PRI swap(value) | swapped
  return value
  swapped :=  value              << 24
  swapped |= (value & $0000ff00) << 8
  swapped |= (value & $00ff0000) >> 8
  swapped |=  value              >> 24
  return swapped

PRI Fatal(msg)
#ifdef TV_DEBUG
    tv.str(string("Error: "))
    tv.str(msg)
    crlf
#endif
    repeat                                             '<== LOOP HERE

PRI CheckError(r)
  if r < 0
#ifdef TV_DEBUG
    tv.str(string("Error: "))
    tv.hex(r,2)
    tv.str(string(" HALTED."))
    crlf
#endif
    repeat                                             '<== LOOP HERE

DAT
                        org     0
'Copy PAR block into COG
boot_zog                mov     zog_params, par
                        mov     address, zog_params
                        add     address, #16
                        rdlong  dispatch_tab_addr, address
                        wrlong  dispatch_tab_final, address 'Update param block with new address
                        add     address, #4
                        rdlong  pasm_addr, address

'Move ZOG dispatch table runtime position in HUB RAM
move_dispatch           mov     src, dispatch_tab_addr
                        mov     dest, dispatch_tab_final
                        mov     count, #$80
:loop                   rdbyte  temp, src
                        wrbyte  temp, dest
                        add     src, #1
                        add     dest, #1
                        djnz    count, #:loop

'Launch Zog in this COG.
go_zog                  cogid   coginit_dest            'Get ID of this COG for restarting
                        mov     temp, pasm_addr         'Set ZOG PASM address for coginit
                        shl     temp, #2                'Shift into bits 17:4 (only high 14 bits required)
                        or      coginit_dest, temp      'Place in dest for coginit

                        mov     temp, zog_params        'Set PAR block address for coginit
                        shl     temp, #16               'Move to bits 31:18 (only high 14 bits required)
                        or      coginit_dest, temp      'Combine PASM addr and PAR addr

                        coginit coginit_dest

#ifdef USE_C3_SPI_MEMORY
board_name              byte    "C3", 0
#endif
#ifdef USE_SDRAM_MEMORY
board_name              byte    "SDRAM", 0
#endif

zog_params              long    0
dispatch_tab_addr       long    0
pasm_addr               long    0
address                 long    0
dest                    long    0
src                     long    0
count                   long    0
temp                    long    0
coginit_dest            long    0

'The final run time position of Zog's dispatch table
dispatch_tab_final      long    zog_dispatch_table

                        fit     $01ff

'The End.

