''*********************************************
''*  Full-Duplex Serial Driver v1.1 Extended  *
''*  (C) 2006 Parallax, Inc.                  *
''*********************************************

CON
  buffer_length = 16                 'can be 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, etc.
  buffer_mask   = buffer_length - 1

  zog_mbox_size = 8*4                'size in bytes

  CR = $0d
  LF = $0a
  
VAR

  long  cog                     'cog flag/id

{
  long  rx_head                 '10 contiguous longs
  long  rx_tail
  long  tx_head
  long  tx_tail
  long  rx_pin
  long  tx_pin
  long  rxtx_mode
  long  bit_ticks
  long  buffer_ptr
  long  zog_mbox
}

PUB start(data, rxpin, txpin, mode, baudrate, buffer, zogmbox) : okay

'' Start serial driver - starts a cog
'' returns false if no cog available
''
'' mode bit 0 = invert rx
'' mode bit 1 = invert tx
'' mode bit 2 = open-drain/source tx
'' mode bit 3 = ignore tx echo on rx

  stop
  longfill(data, 0, 4)
  longmove(data + 4*4, @rxpin, 6)
  long[data][7] := clkfreq / baudrate
  okay := cog := cognew(@entry, data) + 1

PUB stop

'' Stop serial driver - frees a cog

  if cog
    cogstop(cog~ - 1)


DAT

'***********************************
'* Assembly language serial driver *
'***********************************

                        org
'
'
' Entry
'
entry                   mov     t1,par                'get structure address
                        add     t1,#4 << 2            'skip past heads and tails

                        rdlong  t2,t1                 'get rx_pin
                        mov     rxmask,#1
                        shl     rxmask,t2

                        add     t1,#4                 'get tx_pin
                        rdlong  t2,t1
                        mov     txmask,#1
                        shl     txmask,t2

                        add     t1,#4                 'get rxtx_mode
                        rdlong  rxtxmode,t1

                        add     t1,#4                 'get bit_ticks
                        rdlong  bitticks,t1

                        add     t1,#4                 'get buffer_ptr ...
                        rdlong  rxbuff,t1             '... for the receiver
                        mov     txbuff,rxbuff         '... and the transmitter
                        add     txbuff,buffer_length_val

                        add     t1,#4                 'get the zog mailbox ptr
                        rdlong  zog_mbox,t1

                        test    rxtxmode,#%100  wz    'if_nz = open drain Tx
                        test    rxtxmode,#%010  wc    'if_c = inverted output
        if_z_ne_c       or      outa,txmask
        if_z            or      dira,txmask

                        mov     txcode,#transmit      'initialize ping-pong multitasking
                        mov     mbcode,#mbcheck
                        mov     cmdcode,#cmdcheck
                        mov     stopped,#0
'
'
' Receive
'
receive                 jmpret  rxcode,txcode         'run a chunk of transmit code, then return

                        test    rxtxmode,#%001  wz    'wait for start bit on rx pin
                        test    rxmask,ina      wc
        if_z_eq_c       jmp     #receive

                        mov     rxbits,#9             'ready to receive byte
                        mov     rxcnt,bitticks
                        shr     rxcnt,#1              'half a bit tick
                        add     rxcnt,cnt             '+ the current clock             

:bit                    add     rxcnt,bitticks        'ready for the middle of the bit period

:wait                   jmpret  rxcode,txcode         'run a chuck of transmit code, then return

                        mov     t1,rxcnt              'check if bit receive period done
                        sub     t1,cnt
                        cmps    t1,#0           wc
        if_nc           jmp     #:wait

                        test    rxmask,ina      wc    'receive bit on rx pin into carry
                        rcr     rxdata,#1             'shift carry into receiver
                        djnz    rxbits,#:bit          'go get another bit till done

                        test    rxtxmode,#%001  wz    'find out if rx is inverted
        if_z_ne_c       jmp     #receive              'abort if no stop bit

                        shr     rxdata,#32-9          'justify and trim received byte
                        and     rxdata,#$FF
        if_nz           xor     rxdata,#$FF           'if rx inverted, invert byte

                        rdlong  t2,par                'rx_head
                        add     t2,rxbuff             'plus the buffer offset
                        wrbyte  rxdata,t2             'write the byte
                        sub     t2,rxbuff
                        add     t2,#1                 'update rx_head
                        and     t2,buffer_mask_val
                        wrlong  t2,par                'and save

                        jmp     #receive              'byte done, receive next byte
'
'
' Transmit
'
transmit                jmpret  txcode,mbcode         'run a chunk of mailbox code, then return

                        mov     t1,par                'check for head <> tail
                        add     t1,#2 << 2            'tx_head
                        rdlong  t2,t1
                        add     t1,#1 << 2            'tx_tail
                        rdlong  t3,t1
                        cmp     t2,t3           wz
        if_z            jmp     #transmit

                        add     t3,txbuff             'get byte and inc tail
                        rdbyte  txdata,t3
                        sub     t3,txbuff
                        add     t3,#1
                        and     t3,buffer_mask_val
                        wrlong  t3,t1

                        or      txdata,#$100          'or in a stop bit
                        shl     txdata,#2
                        or      txdata,#1             'or in a idle line state and a start bit
                        mov     txbits,#11
                        mov     txcnt,cnt

:bit                    test    rxtxmode,#%100  wz    'output bit on tx pin according to mode
                        test    rxtxmode,#%010  wc
        if_z_and_c      xor     txdata,#1
                        shr     txdata,#1       wc
        if_z            muxc    outa,txmask        
        if_nz           muxnc   dira,txmask
                        add     txcnt,bitticks        'ready next cnt

:wait                   jmpret  txcode,mbcode         'run a chunk of mailbox code, then return

                        mov     t1,txcnt              'check if bit transmit period done
                        sub     t1,cnt
                        cmps    t1,#0           wc
        if_nc           jmp     #:wait

                        djnz    txbits,#:bit          'another bit to transmit?

                        jmp     #transmit             'byte done, transmit next byte

mbcheck                 jmpret  mbcode,cmdcode        'run a chunk of command check code, then return
                        tjnz    stopped,#mbcheck

                        rdlong  t1,zog_mbox     wz    'check for a new command
        if_z            jmp     #mbcheck

                        mov     mbox_ptr,zog_mbox
                        mov     mbox_cnt,#zog_mbox_size

                        mov     mbox_byte,#CR
                        call    #send_byte
                        mov     mbox_byte,#LF
                        call    #send_byte

:next                   jmpret  mbcode,cmdcode        'run a chunk of command check code, then return

                        mov     t1,mbox_ptr
                        xor     t1,#%11
                        rdbyte  mbox_data,t1
                        add     mbox_ptr,#1

                        mov     mbox_byte,mbox_data
                        shr     mbox_byte,#4
                        cmp     mbox_byte,#10   wz,wc
        if_b            add     mbox_byte,#"0"
        if_ae           add     mbox_byte,#"a"-10
                        call    #send_byte

                        mov     mbox_byte,mbox_data
                        and     mbox_byte,#$f
                        cmp     mbox_byte,#10   wz,wc
        if_b            add     mbox_byte,#"0"
        if_ae           add     mbox_byte,#"a"-10
                        call    #send_byte

                        mov     t1,mbox_cnt
                        and     t1,#3
                        cmp     t1,#1           wz
        if_z            mov     mbox_byte,#" "
        if_z            call    #send_byte

                        djnz    mbox_cnt,#:next

                        mov     stopped,#1

                        jmp     #mbcheck

cmdcheck                jmpret  cmdcode,rxcode        'run a chunk of receive code, then return

                        mov     t1,par                'rx_head
                        rdlong  t2,t1
                        add     t1,#1 << 2            'rx_tail
                        rdlong  t3,t1
                        cmp     t2,t3           wz
        if_z            jmp     #cmdcheck

                        add     t3,rxbuff             'get byte and inc tail
                        rdbyte  t4,t3
                        sub     t3,rxbuff
                        add     t3,#1
                        and     t3,buffer_mask_val
                        wrlong  t3,t1

                        cmp     t4,#"c"         wz
        if_z            wrlong  zero,zog_mbox
        if_z            mov     stopped,#0

                        jmp     #cmdcheck


' send the byte in mbox_byte
send_byte               jmpret  mbcode,cmdcode        'run a chunk of command check code, then return

                        mov     t1,par                'check for head+1 <> tail
                        add     t1,#2 << 2            'tx_head
                        rdlong  t2,t1
                        mov     t3,t2                 'tx_head+1
                        add     t3,#1
                        and     t3,buffer_mask_val
                        add     t1,#1 << 2            'tx_tail
                        rdlong  t4,t1
                        cmp     t3,t4           wz
        if_z            jmp     #send_byte

                        add     t2,txbuff             'put byte and inc tail
                        wrbyte  mbox_byte,t2
                        sub     t1,#1 << 2            'tx_head
                        wrlong  t3,t1

send_byte_ret           ret

'
'
' Initialized data
'
'
buffer_length_val       long    buffer_length
buffer_mask_val         long    buffer_mask
zero                    long    0

'
' Uninitialized data
'
t1                      res     1
t2                      res     1
t3                      res     1
t4                      res     1

rxtxmode                res     1
bitticks                res     1

rxmask                  res     1
rxbuff                  res     1
rxdata                  res     1
rxbits                  res     1
rxcnt                   res     1
rxcode                  res     1

txmask                  res     1
txbuff                  res     1
txdata                  res     1
txbits                  res     1
txcnt                   res     1
txcode                  res     1

zog_mbox                res     1
stopped                 res     1
mbox_ptr                res     1
mbox_cnt                res     1
mbox_data               res     1
mbox_byte               res     1
mbcode                  res     1

cmdcode                 res     1

