' this selects the mailbox interface
'#define USE_JCACHED_MEMORY

' select only one of the below
'#define USE_SDRAM_MEMORY        ' John Steven Denson (jazzed)'s SDRAM support
'#define USE_C3_SPI_MEMORY       ' Parallax's C3 SPI flash/SRAM support

' load the serial debug stub
'#define SERIAL_DEBUG

' select debugging output to a TV
#define TV_DEBUG

#ifdef USE_SDRAM_MEMORY
#define XCACHE
#endif

CON
  _clkmode = def#_clkmode
  _xinfreq = def#_xinfreq

  TYPE_FILE_WRITE = 1
  TYPE_FLASH_WRITE = 2
  TYPE_RAM_WRITE = 3
  TYPE_DATA = 4
  TYPE_EOF = 5
  TYPE_FILE_LOAD = 6
  TYPE_RUN = 7

  HDR_SOH = 0
  HDR_PKTN = 1
  HDR_TYPE = 2
  HDR_LEN_HI = 3
  HDR_LEN_LO = 4
  HDR_CHK = 5

  PKTHDRLEN = 6
  PKTMAXLEN = 1024
  PKTCRCLEN = 2

  ' character codes
  SOH = $01    ' start of a packet
  ACK = $06
  NAK = $15
  CR  = $0d

  WRITE_FILE = 1
  WRITE_FLASH = 2
  WRITE_RAM = 3

  hub_memory_size = 32 * 1024
  rendezvous_area_size = 128
  zog_dispatch_table_size = 128

#ifdef USE_C3_SPI_MEMORY
  zpu_memory_size = 64 * 1024
#endif
#ifdef USE_SDRAM_MEMORY
  zpu_memory_size = 32 * 1024 * 1024
#endif
  vm_cache_size = 8 * 1024
  vm_mbox_size = 2 * 4
  zog_mbox_size = 8 * 4
  debug_control_size = 10 * 4
  debug_buffer_size = 16 * 2
  rendezvous_area = hub_memory_size - rendezvous_area_size
  zog_dispatch_table = rendezvous_area - zog_dispatch_table_size
  vm_cache = zog_dispatch_table - vm_cache_size
  vm_mbox = rendezvous_area
  zog_mbox = vm_mbox + vm_mbox_size
  debug_control = zog_mbox + zog_mbox_size
  debug_buffer = debug_control + debug_control_size

OBJ
#ifdef USE_C3_SPI_MEMORY
  def  : "userdefs_c3.spin"
  vm   : "spisram_cache_c3"
#endif
#ifdef USE_SDRAM_MEMORY
  def  : "userdefs_sdram.spin"
  vm   : "SdramCache"
#endif
  zog  : "zog.spin"
  ser  : "FullDuplexSerialExtended"
  sd   : "fsrwFemto_rr001"
  dbg  : "serial_debug"
#ifdef TV_DEBUG
  tv   : "TV_Text"
#endif

VAR
  '8 consecutive longs for Zog's par block.
  long par_zpu_memory_addr
  long par_zpu_memory_sz
  long par_initial_pc
  long par_initial_sp
  long par_dispatch_tab_addr
  long par_pasm_addr
  long par_zog_mbox_addr
  long par_vm_mbox_addr

  long ioControl[2]   ' SD parameters
  byte diskbuff[512]  ' SD sector buffer
  long run_ram_image

  long sd_mounted
  long write_mode
  long load_address

  byte pkt_hdr[PKTHDRLEN]
  byte pkt_data[PKTMAXLEN]
  byte pkt_crc[PKTCRCLEN]
  byte snd_pktn
  byte rcv_pktn

PUB start | type, len

  ser.start(def#conRxPin, def#conTxPin, def#conMode, def#conBaud)
  ser.tx(ACK) ' tell the PC that we're awake

#ifdef TV_DEBUG
  tv.start(def#TvPin)
#ifdef SERIAL_DEBUG
  tv.str(string("Serial Debug Loader v0.2 ("))
#else
  tv.str(string("Serial Loader v0.2 ("))
#endif
  tv.str(@board_name)
  tv.str(string(")", CR))
#endif

  vm.startx(vm_mbox, vm_cache)

  run_ram_image := 0
  sd_mounted := 0

  repeat
    len := rcv_packet(@type, @pkt_data, PKTMAXLEN)
    if len => 0
      case type
        TYPE_FILE_WRITE:        FILE_WRITE_handler
        TYPE_RAM_WRITE:         RAM_WRITE_handler
        TYPE_DATA:              DATA_handler(len)
        TYPE_EOF:               EOF_handler
        TYPE_FILE_LOAD:         FILE_LOAD_handler
        TYPE_RUN:               RUN_handler
#ifdef USE_C3_SPI_MEMORY
        TYPE_FLASH_WRITE:       FLASH_WRITE_handler
#endif
        other:
#ifdef TV_DEBUG
          tv.str(string("Bad packet type: "))
          tv.hex(type, 2)
          crlf
#endif

PRI FILE_WRITE_handler | err
  mountSD
#ifdef TV_DEBUG
  tv.str(string("FILE_WRITE: "))
  tv.str(@pkt_data)
  tv.str(string("...",))
#endif
  err := \sd.popen(@pkt_data, "w")
#ifdef TV_DEBUG
  tv.hex(err, 8)
  crlf
#endif
  write_mode := WRITE_FILE
  load_address := $00000000

PRI RAM_WRITE_handler
#ifdef TV_DEBUG
  tv.str(string("RAM_WRITE", CR))
#endif
  write_mode := WRITE_RAM
  load_address := $00000000
  run_ram_image := 1

PRI DATA_handler(len) | i, val
#ifdef TV_DEBUG
  tv.str(string("DATA: "))
  tv.hex(load_address, 8)
  tv.out(" ")
  tv.dec(len)
  crlf
#endif
  case write_mode
    WRITE_FILE:
      sd.pwrite(@pkt_data, len)
      load_address += len
    WRITE_RAM:
      repeat i from 0 to len - 1 step 4
        val := (pkt_data[i] << 24) | (pkt_data[i+1] << 16) | (pkt_data[i+2] << 8) | pkt_data[i+3]
        vm.writeLong(load_address, val)
        load_address += 4
#ifdef USE_C3_SPI_MEMORY
    WRITE_FLASH:
      vm.WriteFlash(load_address, @pkt_data, len)
      load_address += len
      if (load_address & $00000fff) == 0
        vm.eraseFlashBlock(load_address)
#endif
    other:
#ifdef TV_DEBUG
      tv.str(string("Bad write_mode: "))
      tv.hex(write_mode, 8)
      crlf
#endif

PRI EOF_handler
#ifdef TV_DEBUG
  tv.str(string("EOF", CR))
#endif
  case write_mode
    WRITE_FILE:
      sd.pclose
    WRITE_RAM:
#ifdef USE_C3_SPI_MEMORY
    WRITE_FLASH:
#endif
    other:
#ifdef TV_DEBUG
      tv.str(string("Bad write_mode: "))
      tv.hex(write_mode, 8)
      crlf
#endif

PRI FILE_LOAD_handler
#ifdef TV_DEBUG
  tv.str(string("FILE_LOAD: "))
  tv.str(@pkt_data)
  crlf
#endif
  mountSD
  load_image(@pkt_data)

PRI RUN_handler
#ifdef TV_DEBUG
  tv.str(string("RAM_RUN", CR))
#endif
  if  run_ram_image
    start_bytecode_in_ram
  else
#ifdef USE_C3_SPI_MEMORY
    start_bytecode_in_flash
#endif
#ifdef USE_SDRAM_MEMORY
    Fatal(string("Flash memory not supported"))
#endif
  start_zog

#ifdef USE_C3_SPI_MEMORY

PRI FLASH_WRITE_handler
#ifdef TV_DEBUG
  tv.str(string("FLASH_WRITE", CR))
#endif
  write_mode := WRITE_FLASH
  load_address := $00000000 ' offset into flash
  vm.eraseFlashBlock(load_address)
  run_ram_image := 0

#endif

#ifdef TV_DEBUG
PRI crlf
  tv.out(CR)
#endif

PRI mountSD | err
  if sd_mounted == 0
    if sd.startSD(@ioControl) == 0
      return 0
    repeat
#ifdef TV_DEBUG
      tv.str(string("Mounting SD card...", CR))
#endif
      err := \sd.mount_explicit(def#spiDO,def#spiClk,def#spiDI,def#spiCS)
    until err == 0
    sd_mounted := 1
  return 1

PUB load_image(filename) | bytes, err
#ifdef TV_DEBUG
  tv.str(string("Opening ZPU image: "))
  tv.str(filename)
  tv.str(string("..."))
#endif
  err := \sd.popen(filename, "r")
#ifdef TV_DEBUG
  tv.hex(err, 8)
  crlf
  tv.str(string("Reading image..."))
#endif

  bytes := sd.pread(@diskbuff, 512)
  if bytes < 4
    Fatal(string("Bad image header"))

  if long[@diskbuff] == $00000000
    load_bytecode(bytes)
  else
#ifdef USE_C3_SPI_MEMORY
    write_bytecode_to_flash(bytes)
#endif
#ifdef USE_SDRAM_MEMORY
    Fatal(string("Flash loads not supported"))
#endif

' called with data already in diskbuff
PUB load_bytecode(bytes) | addr, val, i, err
  addr := 0
  err := sd.stopSDCard                                'Stop SD card, but do not stop cog
  CheckError(err)
  repeat while bytes > 0
    repeat i from 0 to bytes - 1 step 4
      val := (diskbuff[i] << 24) | (diskbuff[i+1] << 16) | (diskbuff[i+2] << 8) | diskbuff[i+3]
      vm.writeLong(addr, val)
      addr += 4
    err := \sd.initSDCard(def#spiDO,def#spiClk,def#spiDI,def#spiCS)     'Init the SD
    bytes := sd.pread(@diskbuff, 512)
    err := sd.stopSDCard                                'Stop SD card, but do not stop cog
    CheckError(err)
#ifdef TV_DEBUG
  tv.dec(addr)
  tv.str(string(" bytes"))
  crlf
#endif
  run_ram_image := 1

PUB start_bytecode_in_ram | bss_start, bss_end, i

  ' load the image header
  load_address   := swap(vm.readLong($00000000))
  par_initial_pc := swap(vm.readLong($00000004))
  bss_start      := swap(vm.readLong($00000008))
  bss_end        := swap(vm.readLong($0000000c))

#ifdef TV_DEBUG
  taghex8(string("load_address: "), load_address)
  taghex8(string("start:        "), par_initial_pc)
  taghex8(string("bss_start:    "), bss_start)
  taghex8(string("bss_end:      "), bss_end)
#endif

  ' clear .bss
  repeat while bss_start < bss_end
    vm.writeLong(bss_start, 0)
    bss_start += 4

#ifdef USE_C3_SPI_MEMORY

' called with data already in diskbuff
PUB write_bytecode_to_flash(bytes) | i, addr, err
  addr := 0
  err := sd.stopSDCard                                'Stop SD card, but do not stop cog
  CheckError(err)
  vm.eraseFlashBlock(addr)
  repeat while bytes > 0
    if bytes > 0
      vm.writeFlash(addr, @diskbuff, bytes)
      addr += bytes
      if (addr & $00000fff) == 0
        vm.eraseFlashBlock(addr)
      err := \sd.initSDCard(def#spiDO,def#spiClk,def#spiDI,def#spiCS)     'Init the SD
      bytes := sd.pread(@diskbuff, 512)
      err := sd.stopSDCard                                'Stop SD card, but do not stop cog
      CheckError(err)
#ifdef TV_DEBUG
  tv.dec(addr)
  tv.str(string(" bytes"))
  crlf
#endif
  run_ram_image := 0

PUB start_bytecode_in_flash | bss_start, bss_end, data_image, data_start, data_end

  ' load the image header
  load_address   := swap(vm.readLong($00100000))
  par_initial_pc := swap(vm.readLong($00100004))
  bss_start      := swap(vm.readLong($00100008))
  bss_end        := swap(vm.readLong($0010000c))
  data_image     := swap(vm.readLong($00100010))
  data_start     := swap(vm.readLong($00100014))
  data_end       := swap(vm.readLong($00100018))

#ifdef TV_DEBUG
  taghex8(string("load_address: "), load_address)
  taghex8(string("start:        "), par_initial_pc)
  taghex8(string("bss_start:    "), bss_start)
  taghex8(string("bss_end:      "), bss_end)
  taghex8(string("data_image:   "), data_image)
  taghex8(string("data_start:   "), data_start)
  taghex8(string("data_end:     "), data_end)
#endif

  ' initialize .data
  repeat while data_start < data_end
    vm.writeLong(data_start, vm.readLong(data_image))
    data_image += 4
    data_start += 4

  ' clear .bss
  repeat while bss_start < bss_end
    vm.writeLong(bss_start, 0)
    bss_start += 4

#endif

PRI swap(value) | swapped
  return value
  swapped :=  value              << 24
  swapped |= (value & $0000ff00) << 8
  swapped |= (value & $00ff0000) >> 8
  swapped |=  value              >> 24
  return swapped

#ifdef TV_DEBUG
PRI taghex8(tag, addr)
  tv.str(tag)
  tv.hex(addr, 8)
  crlf
#endif

PUB start_zog | cog

  'Set up Zog's parameter block (except par_initial_pc)
  par_zpu_memory_addr := 0      ' using external memory
  par_zpu_memory_sz := 0        ' using external memory
  par_initial_sp := zpu_memory_size - 8
  par_dispatch_tab_addr := zog.getdispatch_table
  par_pasm_addr := zog.getzog
  par_zog_mbox_addr := zog_mbox
  par_vm_mbox_addr := vm_mbox

  waitcnt(cnt+clkfreq)                               'Give serial terminal window time

  ' stop all COGs except the one running the vm
  ser.stop
  sd.stop

#ifdef SERIAL_DEBUG
  ' start the serial debug COG
  cog := dbg.start(debug_control, def#conRxPin, def#conTxPin, def#conMode, def#conBaud, debug_buffer, zog_mbox)
#ifdef TV_DEBUG
  if cog == 0
    tv.str(string("Failed to start serial debug COG", CR))
    repeat
#endif
#endif

#ifdef TV_DEBUG
  tv.stop
#endif

  ' let ZOG replace the Spin loader code
  coginit(cogid, @boot_zog, @par_zpu_memory_addr)

PRI Fatal(msg)
#ifdef TV_DEBUG
    tv.str(string("Error: "))
    tv.str(msg)
    crlf
#endif
    repeat                                             '<== LOOP HERE

PRI CheckError(r)
  if r < 0
#ifdef TV_DEBUG
    tv.str(string("Error: "))
    tv.hex(r,2)
    tv.str(string(" HALTED."))
    crlf
#endif
    repeat                                             '<== LOOP HERE

{
PRI snd_packet(type, buf, len) | crc16, i

  ' setup the packet header
  pkt_hdr[HDR_SOH] := SOH                                               ' SOH
  pkt_hdr[HDR_PKTN] := ++snd_pktn                                       ' packet#
  pkt_hdr[HDR_TYPE] := type                                             ' type
  pkt_hdr[HDR_LEN_HI] := len >> 8                                       ' data length - high byte
  pkt_hdr[HDR_LEN_LO] := len                                            ' data length - low byte
  pkt_hdr[HDR_CHK] := pkt_hdr[1] + pkt_hdr[2] + pkt_hdr[3] + pkt_hdr[4] ' header checksum

  ' compute the crc
  crc16 := 0
  repeat i from 0 to len - 1
    crc16 := updcrc(crc16, buf[i])
  crc16 := updcrc(crc16, 0)
  crc16 := updcrc(crc16, 0)

  ' add the crc to the frame
  pkt_crc[0] := crc16 >> 8
  pkt_crc[1] := crc16

  ' send the packet
  txbuf(@pkt_hdr, PKTHDRLEN)
  txbuf(@buf, len)
  txbuf(@pkt_crc, PKTCRCLEN)

PRI txbuf(buf, len) | i
  repeat i from 0 to len - 1
    ser.tx(byte[buf][i])
}

PRI rcv_packet(ptype, buf, len) | actual_len, crc16, chk, ch, i

  ' look for start of packet */
  repeat
    pkt_hdr[HDR_SOH] := ser.rx
  until pkt_hdr[HDR_SOH] == SOH

  ' receive the rest of the header
  repeat i from 1 to PKTHDRLEN - 1
    pkt_hdr[i] := ser.rx

  ' check the header checksum
  chk := (pkt_hdr[1] + pkt_hdr[2] + pkt_hdr[3] + pkt_hdr[4]) & $ff
  if pkt_hdr[HDR_CHK] <> chk
#ifdef TV_DEBUG
    tv.str(string("Header checksum error: "))
    tv.hex(pkt_hdr[HDR_CHK], 2)
    tv.str(string(" "))
    tv.hex(chk, 2)
    crlf
#endif
    ser.tx(NAK)
    return -1

  ' make sure the buffer is big enough for the payload
  actual_len := (pkt_hdr[HDR_LEN_HI] << 8) | pkt_hdr[HDR_LEN_LO]
  if actual_len > len
#ifdef TV_DEBUG
    tv.str(string("Packet length error", CR))
#endif
    ser.tx(NAK)
    return -1

  ' receive the packet payload
  crc16 := 0
  if actual_len > 0
    repeat i from 0 to actual_len - 1
      ch := ser.rx
      byte[buf][i] := ch
      crc16 := updcrc(crc16, ch)

  ' receive the crc
  pkt_crc[0] := ser.rx
  pkt_crc[1] := ser.rx

  ' check the crc
  crc16 := updcrc(crc16, pkt_crc[0])
  crc16 := updcrc(crc16, pkt_crc[1])
  if crc16 <> 0
#ifdef TV_DEBUG
    tv.str(string("CRC error", CR))
#endif
    ser.tx(NAK)
    return -1

  ' acknowledge the packet
  ser.tx(ACK)

  ' return packet type and the length of the payload
  long[ptype] := pkt_hdr[HDR_TYPE]
  return actual_len

PRI updcrc(crc, data)
  return (word[@crctab][(crc >> 8) & $ff] ^ (crc << 8) ^ data) & $ffff

DAT
                        org     0
'Copy PAR block into COG
boot_zog                mov     zog_params, par
                        mov     address, zog_params
                        add     address, #16
                        rdlong  dispatch_tab_addr, address
                        wrlong  dispatch_tab_final, address 'Update param block with new address
                        add     address, #4
                        rdlong  pasm_addr, address

'Move ZOG dispatch table runtime position in HUB RAM
move_dispatch           mov     src, dispatch_tab_addr
                        mov     dest, dispatch_tab_final
                        mov     count, #$80
:loop                   rdbyte  temp, src
                        wrbyte  temp, dest
                        add     src, #1
                        add     dest, #1
                        djnz    count, #:loop

'Launch Zog in this COG.
go_zog                  cogid   coginit_dest            'Get ID of this COG for restarting
                        mov     temp, pasm_addr         'Set ZOG PASM address for coginit
                        shl     temp, #2                'Shift into bits 17:4 (only high 14 bits required)
                        or      coginit_dest, temp      'Place in dest for coginit

                        mov     temp, zog_params        'Set PAR block address for coginit
                        shl     temp, #16               'Move to bits 31:18 (only high 14 bits required)
                        or      coginit_dest, temp      'Combine PASM addr and PAR addr

                        coginit coginit_dest

#ifdef USE_C3_SPI_MEMORY
board_name              byte    "C3", 0
#endif
#ifdef USE_SDRAM_MEMORY
board_name              byte    "SDRAM", 0
#endif

zog_params              long    0
dispatch_tab_addr       long    0
pasm_addr               long    0
address                 long    0
dest                    long    0
src                     long    0
count                   long    0
temp                    long    0
coginit_dest            long    0

'The final run time position of Zog's dispatch table
dispatch_tab_final      long    zog_dispatch_table

                        fit     $01ff

crctab
    word $0000,  $1021,  $2042,  $3063,  $4084,  $50a5,  $60c6,  $70e7
    word $8108,  $9129,  $a14a,  $b16b,  $c18c,  $d1ad,  $e1ce,  $f1ef
    word $1231,  $0210,  $3273,  $2252,  $52b5,  $4294,  $72f7,  $62d6
    word $9339,  $8318,  $b37b,  $a35a,  $d3bd,  $c39c,  $f3ff,  $e3de
    word $2462,  $3443,  $0420,  $1401,  $64e6,  $74c7,  $44a4,  $5485
    word $a56a,  $b54b,  $8528,  $9509,  $e5ee,  $f5cf,  $c5ac,  $d58d
    word $3653,  $2672,  $1611,  $0630,  $76d7,  $66f6,  $5695,  $46b4
    word $b75b,  $a77a,  $9719,  $8738,  $f7df,  $e7fe,  $d79d,  $c7bc
    word $48c4,  $58e5,  $6886,  $78a7,  $0840,  $1861,  $2802,  $3823
    word $c9cc,  $d9ed,  $e98e,  $f9af,  $8948,  $9969,  $a90a,  $b92b
    word $5af5,  $4ad4,  $7ab7,  $6a96,  $1a71,  $0a50,  $3a33,  $2a12
    word $dbfd,  $cbdc,  $fbbf,  $eb9e,  $9b79,  $8b58,  $bb3b,  $ab1a
    word $6ca6,  $7c87,  $4ce4,  $5cc5,  $2c22,  $3c03,  $0c60,  $1c41
    word $edae,  $fd8f,  $cdec,  $ddcd,  $ad2a,  $bd0b,  $8d68,  $9d49
    word $7e97,  $6eb6,  $5ed5,  $4ef4,  $3e13,  $2e32,  $1e51,  $0e70
    word $ff9f,  $efbe,  $dfdd,  $cffc,  $bf1b,  $af3a,  $9f59,  $8f78
    word $9188,  $81a9,  $b1ca,  $a1eb,  $d10c,  $c12d,  $f14e,  $e16f
    word $1080,  $00a1,  $30c2,  $20e3,  $5004,  $4025,  $7046,  $6067
    word $83b9,  $9398,  $a3fb,  $b3da,  $c33d,  $d31c,  $e37f,  $f35e
    word $02b1,  $1290,  $22f3,  $32d2,  $4235,  $5214,  $6277,  $7256
    word $b5ea,  $a5cb,  $95a8,  $8589,  $f56e,  $e54f,  $d52c,  $c50d
    word $34e2,  $24c3,  $14a0,  $0481,  $7466,  $6447,  $5424,  $4405
    word $a7db,  $b7fa,  $8799,  $97b8,  $e75f,  $f77e,  $c71d,  $d73c
    word $26d3,  $36f2,  $0691,  $16b0,  $6657,  $7676,  $4615,  $5634
    word $d94c,  $c96d,  $f90e,  $e92f,  $99c8,  $89e9,  $b98a,  $a9ab
    word $5844,  $4865,  $7806,  $6827,  $18c0,  $08e1,  $3882,  $28a3
    word $cb7d,  $db5c,  $eb3f,  $fb1e,  $8bf9,  $9bd8,  $abbb,  $bb9a
    word $4a75,  $5a54,  $6a37,  $7a16,  $0af1,  $1ad0,  $2ab3,  $3a92
    word $fd2e,  $ed0f,  $dd6c,  $cd4d,  $bdaa,  $ad8b,  $9de8,  $8dc9
    word $7c26,  $6c07,  $5c64,  $4c45,  $3ca2,  $2c83,  $1ce0,  $0cc1
    word $ef1f,  $ff3e,  $cf5d,  $df7c,  $af9b,  $bfba,  $8fd9,  $9ff8
    word $6e17,  $7e36,  $4e55,  $5e74,  $2e93,  $3eb2,  $0ed1,  $1ef0

'The End.

